"""
Common utility functions for the nf-core python package.
"""

import concurrent.futures
import datetime
import errno
import fnmatch
import hashlib
import io
import json
import logging
import mimetypes
import os
import random
import re
import shlex
import subprocess
import sys
import time
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, Generator, List, Literal, Optional, Tuple, Union

import git
import prompt_toolkit.styles
import questionary
import requests.auth
import requests_cache
import rich
import rich.markup
import yaml
from packaging.version import Version
from pydantic import BaseModel, ValidationError, field_validator
from rich.live import Live
from rich.spinner import Spinner

import nf_core

if TYPE_CHECKING:
    from nf_core.pipelines.schema import PipelineSchema

log = logging.getLogger(__name__)

# ASCII nf-core logo
nfcore_logo = [
    r"[green]                                          ,--.[grey39]/[green],-.",
    r"[blue]          ___     __   __   __   ___     [green]/,-._.--~\ ",
    r"[blue]    |\ | |__  __ /  ` /  \ |__) |__      [yellow]   }  {",
    r"[blue]    | \| |       \__, \__/ |  \ |___     [green]\`-._,-`-,",
    r"[green]                                          `._,._,'",
]

# Custom style for questionary
nfcore_question_style = prompt_toolkit.styles.Style(
    [
        ("qmark", "fg:ansiblue bold"),  # token in front of the question
        ("question", "bold"),  # question text
        (
            "answer",
            "fg:ansigreen nobold bg:",
        ),  # submitted answer text behind the question
        (
            "pointer",
            "fg:ansiyellow bold",
        ),  # pointer used in select and checkbox prompts
        (
            "highlighted",
            "fg:ansiblue bold",
        ),  # pointed-at choice in select and checkbox prompts
        (
            "selected",
            "fg:ansiyellow noreverse bold",
        ),  # style for a selected item of a checkbox
        ("separator", "fg:ansiblack"),  # separator in lists
        ("instruction", ""),  # user instructions for select, rawselect, checkbox
        ("text", ""),  # plain text
        (
            "disabled",
            "fg:gray italic",
        ),  # disabled choices for select and checkbox prompts
        ("choice-default", "fg:ansiblack"),
        ("choice-default-changed", "fg:ansiyellow"),
        ("choice-required", "fg:ansired"),
    ]
)

NFCORE_CACHE_DIR = Path(
    os.environ.get("XDG_CACHE_HOME", Path(os.getenv("HOME") or "", ".cache")),
    "nfcore",
)
NFCORE_DIR = Path(os.environ.get("XDG_CONFIG_HOME", os.path.join(os.getenv("HOME") or "", ".config")), "nfcore")


def fetch_remote_version(source_url):
    response = requests.get(source_url, timeout=3)
    remote_version = re.sub(r"[^0-9\.]", "", response.text)
    return remote_version


def check_if_outdated(
    current_version=None,
    remote_version=None,
    source_url="https://nf-co.re/tools_version",
):
    """
    Check if the current version of nf-core is outdated
    """
    # Exit immediately if disabled via ENV var
    if os.environ.get("NFCORE_NO_VERSION_CHECK", False):
        return (True, "", "")
    # Set and clean up the current version string
    if current_version is None:
        current_version = nf_core.__version__
    current_version = re.sub(r"[^0-9\.]", "", current_version)
    # Build the URL to check against
    source_url = os.environ.get("NFCORE_VERSION_URL", source_url)
    source_url = f"{source_url}?v={current_version}"
    # check if we have a newer version without blocking the rest of the script
    is_outdated = False
    if remote_version is None:  # we set it manually for tests
        try:
            with concurrent.futures.ThreadPoolExecutor() as executor:
                future = executor.submit(fetch_remote_version, source_url)
                remote_version = future.result()
        except Exception as e:
            log.debug(f"Could not check for nf-core updates: {e}")
    if remote_version is not None:
        if Version(remote_version) > Version(current_version):
            is_outdated = True
    return (is_outdated, current_version, remote_version)


def rich_force_colors():
    """
    Check if any environment variables are set to force Rich to use coloured output
    """
    if os.getenv("GITHUB_ACTIONS") or os.getenv("FORCE_COLOR") or os.getenv("PY_COLORS"):
        return True
    return None


class Pipeline:
    """Object to hold information about a local pipeline.

    Args:
        path (str): The path to the nf-core pipeline directory.

    Attributes:
        conda_config (dict): The parsed conda configuration file content (``environment.yml``).
        conda_package_info (dict): The conda package(s) information, based on the API requests to Anaconda cloud.
        nf_config (dict): The Nextflow pipeline configuration file content.
        files (list): A list of files found during the linting process.
        git_sha (str): The git sha for the repo commit / current GitHub pull-request (`$GITHUB_PR_COMMIT`)
        minNextflowVersion (str): The minimum required Nextflow version to run the pipeline.
        wf_path (str): Path to the pipeline directory.
        pipeline_name (str): The pipeline name, without the `nf-core` tag, for example `hlatyping`.
        schema_obj (obj): A :class:`PipelineSchema` object
    """

    def __init__(self, wf_path: Path) -> None:
        """Initialise pipeline object"""
        self.conda_config: Dict = {}
        self.conda_package_info: Dict = {}
        self.nf_config: Dict = {}
        self.files: List[Path] = []
        self.git_sha: Optional[str] = None
        self.minNextflowVersion: Optional[str] = None
        self.wf_path = Path(wf_path)
        self.pipeline_name: Optional[str] = None
        self.pipeline_prefix: Optional[str] = None
        self.schema_obj: Optional[PipelineSchema] = None
        self.repo: Optional[git.Repo] = None

        try:
            self.repo = git.Repo(self.wf_path)
            self.git_sha = self.repo.head.object.hexsha
        except Exception as e:
            log.debug(f"Could not find git hash for pipeline: {self.wf_path}. {e}")

        # Overwrite if we have the last commit from the PR - otherwise we get a merge commit hash
        if os.environ.get("GITHUB_PR_COMMIT", "") != "":
            self.git_sha = os.environ["GITHUB_PR_COMMIT"]

    def __repr__(self) -> str:
        return f"<Pipeline '{self.pipeline_name}' at {self.wf_path}>"

    def _load(self) -> bool:
        """Run core load functions"""

        return self.load_pipeline_config() and self._load_conda_environment()

    def _load_conda_environment(self) -> bool:
        """Try to load the pipeline environment.yml file, if it exists"""
        try:
            with open(Path(self.wf_path, "environment.yml")) as fh:
                self.conda_config = yaml.safe_load(fh)
            return True
        except FileNotFoundError:
            log.debug("No conda `environment.yml` file found.")
            return False

    def _fp(self, fn: Union[str, Path]) -> Path:
        """Convenience function to get full path to a file in the pipeline"""
        return Path(self.wf_path, fn)

    def list_files(self) -> List[Path]:
        """Get a list of all files in the pipeline"""
        files = []
        try:
            # First, try to get the list of files using git
            git_ls_files = subprocess.check_output(["git", "ls-files"], cwd=self.wf_path).splitlines()
            for fn in git_ls_files:
                full_fn = Path(self.wf_path) / fn.decode("utf-8")
                if full_fn.is_file():
                    files.append(full_fn)
                else:
                    log.debug(f"`git ls-files` returned '{full_fn}' but could not open it!")
        except subprocess.CalledProcessError:
            # Failed, so probably not initialised as a git repository - just a list of all files
            files = []
            for file_path in self.wf_path.rglob("*"):
                if file_path.is_file():
                    # Append the file path to the list
                    files.append(file_path)
            if len(files) == 0:
                log.debug(f"No files found in pipeline: {self.wf_path}")

        return files

    def load_pipeline_config(self) -> bool:
        """Get the nextflow config for this pipeline

        Once loaded, set a few convenience reference class attributes
        """
        self.nf_config = fetch_wf_config(self.wf_path)

        self.pipeline_prefix, self.pipeline_name = self.nf_config.get("manifest.name", "/").strip("'").split("/")

        nextflow_version_match = re.search(r"[0-9\.]+(-edge)?", self.nf_config.get("manifest.nextflowVersion", ""))
        if nextflow_version_match:
            self.minNextflowVersion = nextflow_version_match.group(0)
            return True
        return False


def is_pipeline_directory(wf_path):
    """
    Checks if the specified directory have the minimum required files
    ('main.nf', 'nextflow.config') for a pipeline directory

    Args:
        wf_path (str): The directory to be inspected

    Raises:
        UserWarning: If one of the files are missing
    """
    for fn in ["main.nf", "nextflow.config"]:
        path = os.path.join(wf_path, fn)
        if not os.path.isfile(path):
            if wf_path == ".":
                warning = f"Current directory is not a pipeline - '{fn}' is missing."
            else:
                warning = f"'{wf_path}' is not a pipeline - '{fn}' is missing."
            raise UserWarning(warning)


def fetch_wf_config(wf_path: Path, cache_config: bool = True) -> dict:
    """Uses Nextflow to retrieve the the configuration variables
    from a Nextflow workflow.

    Args:
        wf_path (str): Nextflow workflow file system path.
        cache_config (bool): cache configuration or not (def. True)

    Returns:
        dict: Workflow configuration settings.
    """

    log.debug(f"Got '{wf_path}' as path")
    wf_path = Path(wf_path)
    config = {}
    cache_fn = None
    cache_basedir = None
    cache_path = None

    # Nextflow home directory - use env var if set, or default to ~/.nextflow
    nxf_home = Path(os.environ.get("NXF_HOME", Path(os.getenv("HOME") or "", ".nextflow")))

    # Build a cache directory if we can
    if nxf_home.is_dir():
        cache_basedir = Path(nxf_home, "nf-core")
        if not cache_basedir.is_dir():
            cache_basedir.mkdir(parents=True, exist_ok=True)

    # If we're given a workflow object with a commit, see if we have a cached copy
    cache_fn = None
    # Make a filename based on file contents
    concat_hash = ""
    for fn in ["nextflow.config", "main.nf"]:
        try:
            with open(wf_path / fn, "rb") as fh:
                concat_hash += hashlib.sha256(fh.read()).hexdigest()
        except FileNotFoundError:
            pass
    # Hash the hash
    if len(concat_hash) > 0:
        bighash = hashlib.sha256(concat_hash.encode("utf-8")).hexdigest()
        cache_fn = f"wf-config-cache-{bighash[:25]}.json"

    if cache_basedir and cache_fn:
        cache_path = Path(cache_basedir, cache_fn)
        if cache_path.is_file() and cache_config is True:
            log.debug(f"Found a config cache, loading: {cache_path}")
            with open(cache_path) as fh:
                try:
                    config = json.load(fh)
                except json.JSONDecodeError as e:
                    raise UserWarning(f"Unable to load JSON file '{cache_path}' due to error {e}")
            return config
    log.debug("No config cache found")

    # Call `nextflow config`
    result = run_cmd("nextflow", f"config -flat {wf_path}")
    if result is not None:
        nfconfig_raw, _ = result
        for line in nfconfig_raw.splitlines():
            ul = line.decode("utf-8")
            try:
                k, v = ul.split(" = ", 1)
                config[k] = v.strip("'\"")
            except ValueError:
                log.debug(f"Couldn't find key=value config pair:\n  {ul}")

    # Scrape main.nf for additional parameter declarations
    # Values in this file are likely to be complex, so don't both trying to capture them. Just get the param name.
    try:
        main_nf = Path(wf_path, "main.nf")
        with open(main_nf, "rb") as fh:
            for line in fh:
                line_str = line.decode("utf-8")
                match = re.match(r"^\s*(params\.[a-zA-Z0-9_]+)\s*=(?!=)", line_str)
                if match:
                    config[match.group(1)] = "null"
    except FileNotFoundError as e:
        log.debug(f"Could not open {main_nf} to look for parameter declarations - {e}")

    # If we can, save a cached copy
    # HINT: during testing phase (in test_download, for example) we don't want
    # to save configuration copy in $HOME, otherwise the tests/pipelines/test_download.py::DownloadTest::test_wf_use_local_configs
    # will fail after the first attempt. It's better to not save temporary data
    # in others folders than tmp when doing tests in general
    if cache_path and cache_config:
        log.debug(f"Saving config cache: {cache_path}")
        with open(cache_path, "w") as fh:
            json.dump(config, fh, indent=4)

    return config


def run_cmd(executable: str, cmd: str) -> Union[Tuple[bytes, bytes], None]:
    """Run a specified command and capture the output. Handle errors nicely."""
    full_cmd = f"{executable} {cmd}"
    log.debug(f"Running command: {full_cmd}")
    try:
        proc = subprocess.run(shlex.split(full_cmd), capture_output=True, check=True)
        return (proc.stdout, proc.stderr)
    except OSError as e:
        if e.errno == errno.ENOENT:
            raise RuntimeError(
                f"It looks like {executable} is not installed. Please ensure it is available in your PATH."
            )
        else:
            return None
    except subprocess.CalledProcessError as e:
        log.debug(f"Command '{full_cmd}' returned non-zero error code '{e.returncode}':\n[red]> {e.stderr.decode()}")
        if executable == "nf-test":
            return (e.stdout, e.stderr)
        else:
            raise RuntimeError(
                f"Command '{full_cmd}' returned non-zero error code '{e.returncode}':\n[red]> {e.stderr.decode()}{e.stdout.decode()}"
            )


def setup_nfcore_dir() -> bool:
    """Creates a directory for files that need to be kept between sessions

    Currently only used for keeping local copies of modules repos
    """
    if not NFCORE_DIR.exists():
        NFCORE_DIR.mkdir(parents=True)
    return True


def setup_requests_cachedir() -> Dict[str, Union[Path, datetime.timedelta, str]]:
    """Sets up local caching for faster remote HTTP requests.

    Caching directory will be set up in the user's home directory under
    a .config/nf-core/cache_* subdir.

    Uses requests_cache monkey patching.
    Also returns the config dict so that we can use the same setup with a Session.
    """
    pyversion: str = ".".join(str(v) for v in sys.version_info[0:3])
    cachedir: Path = setup_nfcore_cachedir(f"cache_{pyversion}")
    config: Dict[str, Union[Path, datetime.timedelta, str]] = {
        "cache_name": Path(cachedir, "github_info"),
        "expire_after": datetime.timedelta(hours=1),
        "backend": "sqlite",
    }

    logging.getLogger("requests_cache").setLevel(logging.WARNING)
    return config


def setup_nfcore_cachedir(cache_fn: Union[str, Path]) -> Path:
    """Sets up local caching for caching files between sessions."""

    cachedir = Path(NFCORE_CACHE_DIR, cache_fn)

    try:
        if not Path(cachedir).exists():
            Path(cachedir).mkdir(parents=True)
    except PermissionError:
        log.warn(f"Could not create cache directory: {cachedir}")

    return cachedir


def wait_cli_function(poll_func: Callable[[], bool], refresh_per_second: int = 20) -> None:
    """
    Display a command-line spinner while calling a function repeatedly.

    Keep waiting until that function returns True

    Arguments:
       poll_func (function): Function to call
       refresh_per_second (int): Refresh this many times per second. Default: 20.

    Returns:
       None. Just sits in an infinite loop until the function returns True.
    """
    try:
        spinner = Spinner("dots2", "Use ctrl+c to stop waiting and force exit.")
        with Live(spinner, refresh_per_second=refresh_per_second):
            while True:
                if poll_func():
                    break
                time.sleep(2)
    except KeyboardInterrupt:
        raise AssertionError("Cancelled!")


def poll_nfcore_web_api(api_url: str, post_data: Optional[Dict] = None) -> Dict:
    """
    Poll the nf-core website API

    Takes argument api_url for URL

    Expects API response to be valid JSON and contain a top-level 'status' key.
    """
    # Run without requests_cache so that we get the updated statuses
    with requests_cache.disabled():
        try:
            if post_data is None:
                response = requests.get(api_url, headers={"Cache-Control": "no-cache"})
            else:
                log.debug(f"requesting {api_url} with {post_data}")
                response = requests.post(url=api_url, data=post_data)
        except requests.exceptions.Timeout:
            raise AssertionError(f"URL timed out: {api_url}")
        except requests.exceptions.ConnectionError:
            raise AssertionError(f"Could not connect to URL: {api_url}")
        else:
            if response.status_code != 200 and response.status_code != 301:
                response_content = response.content
                if isinstance(response_content, bytes):
                    response_content = response_content.decode()
                log.debug(f"Response content:\n{response_content}")
                raise AssertionError(
                    f"Could not access remote API results: {api_url} (HTML {response.status_code} Error)"
                )
            # follow redirects
            if response.status_code == 301:
                return poll_nfcore_web_api(response.headers["Location"], post_data)
            try:
                web_response = json.loads(response.content)
                if "status" not in web_response:
                    raise AssertionError()
            except (json.decoder.JSONDecodeError, AssertionError, TypeError):
                response_content = response.content
                if isinstance(response_content, bytes):
                    response_content = response_content.decode()
                log.debug(f"Response content:\n{response_content}")
                raise AssertionError(
                    f"nf-core website API results response not recognised: {api_url}\n "
                    "See verbose log for full response"
                )
            else:
                return web_response


class GitHubAPISession(requests_cache.CachedSession):
    """
    Class to provide a single session for interacting with the GitHub API for a run.
    Inherits the requests_cache.CachedSession and adds additional functionality,
    such as automatically setting up GitHub authentication if we can.
    """

    def __init__(self) -> None:
        self.auth_mode: Optional[str] = None
        self.return_ok: List[int] = [200, 201]
        self.return_retry: List[int] = [403]
        self.return_unauthorised: List[int] = [401]
        self.has_init: bool = False

    def lazy_init(self) -> None:
        """
        Initialise the object.

        Only do this when it's actually being used (due to global import)
        """
        log.debug("Initialising GitHub API requests session")
        cache_config = setup_requests_cachedir()
        super().__init__(**cache_config)
        self.setup_github_auth()
        self.has_init = True

    def setup_github_auth(self, auth=None):
        """
        Try to automatically set up GitHub authentication
        """
        if auth is not None:
            self.auth = auth
            self.auth_mode = "supplied to function"

        # Class for Bearer token authentication
        # https://stackoverflow.com/a/58055668/713980
        class BearerAuth(requests.auth.AuthBase):
            def __init__(self, token):
                self.token = token

            def __call__(self, r):
                r.headers["authorization"] = f"Bearer {self.token}"
                return r

        # Default auth if we're running and the gh CLI tool is installed
        gh_cli_config_fn = os.path.expanduser("~/.config/gh/hosts.yml")
        if self.auth is None and os.path.exists(gh_cli_config_fn):
            try:
                with open(gh_cli_config_fn) as fh:
                    gh_cli_config = yaml.safe_load(fh)
                    self.auth = requests.auth.HTTPBasicAuth(
                        gh_cli_config["github.com"]["user"],
                        gh_cli_config["github.com"]["oauth_token"],
                    )
                    self.auth_mode = f"gh CLI config: {gh_cli_config['github.com']['user']}"
            except Exception:
                ex_type, ex_value, _ = sys.exc_info()
                if ex_type is not None:
                    output = rich.markup.escape(f"{ex_type.__name__}: {ex_value}")
                    log.debug(f"Couldn't auto-auth with GitHub CLI auth from '{gh_cli_config_fn}': [red]{output}")

        # Default auth if we have a GitHub Token (eg. GitHub Actions CI)
        if os.environ.get("GITHUB_TOKEN") is not None and self.auth is None:
            self.auth_mode = "Bearer token with GITHUB_TOKEN"
            self.auth = BearerAuth(os.environ["GITHUB_TOKEN"])
        else:
            log.warning("Could not find GitHub authentication token. Some API requests may fail.")

        log.debug(f"Using GitHub auth: {self.auth_mode}")

    def log_content_headers(self, request, post_data=None):
        """
        Try to dump everything to the console, useful when things go wrong.
        """
        log.debug(f"Requested URL: {request.url}")
        log.debug(f"From requests cache: {request.from_cache}")
        log.debug(f"Request status code: {request.status_code}")
        log.debug(f"Request reason: {request.reason}")
        if post_data is None:
            post_data = {}
        try:
            log.debug(json.dumps(dict(request.headers), indent=4))
            log.debug(json.dumps(request.json(), indent=4))
            log.debug(json.dumps(post_data, indent=4))
        except Exception as e:
            log.debug(f"Could not parse JSON response from GitHub API! {e}")
            log.debug(request.headers)
            log.debug(request.content)
            log.debug(post_data)

    def safe_get(self, url):
        """
        Run a GET request, raise a nice exception with lots of logging if it fails.
        """
        if not self.has_init:
            self.lazy_init()
        request = self.get(url)
        if request.status_code in self.return_retry:
            stderr = rich.console.Console(stderr=True, force_terminal=rich_force_colors())
            try:
                r = self.request_retry(url)
            except Exception as e:
                stderr.print_exception()
                raise e
            else:
                return r
        elif request.status_code in self.return_unauthorised:
            raise RuntimeError("GitHub API PR failed, probably due to an expired GITHUB_TOKEN.")

        return request

    def get(self, url, **kwargs):
        """
        Initialise the session if we haven't already, then call the superclass get method.
        """
        if not self.has_init:
            self.lazy_init()
        return super().get(url, **kwargs)

    def request_retry(self, url, post_data=None):
        """
        Try to fetch a URL, keep retrying if we get a certain return code.

        Used in nf-core pipelines sync code because we get 403 errors: too many simultaneous requests
        See https://github.com/nf-core/tools/issues/911
        """
        if not self.has_init:
            self.lazy_init()

        # Start the loop for a retry mechanism
        while True:
            # GET request
            if post_data is None:
                log.debug(f"Sending GET request to {url}")
                r = self.get(url=url)
            # POST request
            else:
                log.debug(f"Sending POST request to {url}")
                r = self.post(url=url, json=post_data)

            # Failed but expected - try again
            if r.status_code in self.return_retry:
                self.log_content_headers(r, post_data)
                log.debug(f"GitHub API PR failed - got return code {r.status_code}")
                wait_time = float(re.sub("[^0-9]", "", str(r.headers.get("Retry-After", 0))))
                if wait_time == 0:
                    log.debug("Couldn't find 'Retry-After' header, guessing a length of time to wait")
                    wait_time = random.randrange(10, 60)
                log.warning(f"Got API return code {r.status_code}. Trying again after {wait_time} seconds..")
                time.sleep(wait_time)

            # Unexpected error - raise
            elif r.status_code not in self.return_ok:
                self.log_content_headers(r, post_data)
                raise RuntimeError(f"GitHub API PR failed - got return code {r.status_code} from {url}")

            # Success!
            else:
                return r


# Single session object to use for entire codebase. Not sure if there's a better way to do this?
gh_api = GitHubAPISession()


def anaconda_package(dep, dep_channels=None):
    """Query conda package information.

    Sends a HTTP GET request to the Anaconda remote API.

    Args:
        dep (str): A conda package name.
        dep_channels (list): list of conda channels to use

    Raises:
        A LookupError, if the connection fails or times out or gives an unexpected status code
        A ValueError, if the package name can not be found (404)
    """

    if dep_channels is None:
        dep_channels = ["conda-forge", "bioconda"]

    # Check if each dependency is the latest available version
    if "=" in dep:
        depname, _ = dep.split("=", 1)
    else:
        depname = dep

    # 'defaults' isn't actually a channel name. See https://docs.anaconda.com/anaconda/user-guide/tasks/using-repositories/
    if "defaults" in dep_channels:
        dep_channels.remove("defaults")
        dep_channels.extend(["main", "anaconda", "r", "free", "archive", "anaconda-extras"])
    if "::" in depname:
        dep_channels = [depname.split("::")[0]]
        depname = depname.split("::")[1]

    for ch in dep_channels:
        anaconda_api_url = f"https://api.anaconda.org/package/{ch}/{depname}"
        try:
            response = requests.get(anaconda_api_url, timeout=10)
        except requests.exceptions.Timeout:
            raise LookupError(f"Anaconda API timed out: {anaconda_api_url}")
        except requests.exceptions.ConnectionError:
            raise LookupError("Could not connect to Anaconda API")
        else:
            if response.status_code == 200:
                return response.json()
            if response.status_code != 404:
                raise LookupError(
                    f"Anaconda API returned unexpected response code `{response.status_code}` for: "
                    f"{anaconda_api_url}\n{response}"
                )
            # response.status_code == 404
            log.debug(f"Could not find `{dep}` in conda channel `{ch}`")

    # We have looped through each channel and had a 404 response code on everything
    raise ValueError(f"Could not find Conda dependency using the Anaconda API: '{dep}'")


def parse_anaconda_licence(anaconda_response, version=None):
    """Given a response from the anaconda API using anaconda_package, parse the software licences.

    Returns: Set of licence types
    """
    licences = set()
    # Licence for each version
    for f in anaconda_response["files"]:
        if not version or version == f.get("version"):
            try:
                licences.add(f["attrs"]["license"])
            except KeyError:
                pass
    # Main licence field
    if len(list(licences)) == 0 and isinstance(anaconda_response["license"], str):
        licences.add(anaconda_response["license"])

    # Clean up / standardise licence names
    clean_licences = []
    for license in licences:
        license = re.sub(r"GNU General Public License v\d \(([^\)]+)\)", r"\1", license)
        license = re.sub(r"GNU GENERAL PUBLIC LICENSE", "GPL", license, flags=re.IGNORECASE)
        license = license.replace("GPL-", "GPLv")
        license = re.sub(r"GPL\s*([\d\.]+)", r"GPL v\1", license)  # Add v prefix to GPL version if none found
        license = re.sub(r"GPL\s*v(\d).0", r"GPL v\1", license)  # Remove superfluous .0 from GPL version
        license = re.sub(r"GPL \(([^\)]+)\)", r"GPL \1", license)
        license = re.sub(r"GPL\s*v", "GPL v", license)  # Normalise whitespace to one space between GPL and v
        license = re.sub(r"\s*(>=?)\s*(\d)", r" \1\2", license)  # Normalise whitespace around >= GPL versions
        license = license.replace("Clause", "clause")  # BSD capitalisation
        license = re.sub(r"-only$", "", license)  # Remove superfluous GPL "only" version suffixes
        clean_licences.append(license)
    return clean_licences


def pip_package(dep):
    """Query PyPI package information.

    Sends a HTTP GET request to the PyPI remote API.

    Args:
        dep (str): A PyPI package name.

    Raises:
        A LookupError, if the connection fails or times out
        A ValueError, if the package name can not be found
    """
    pip_depname, _ = dep.split("=", 1)
    pip_api_url = f"https://pypi.python.org/pypi/{pip_depname}/json"
    try:
        response = requests.get(pip_api_url, timeout=10)
    except requests.exceptions.Timeout:
        raise LookupError(f"PyPI API timed out: {pip_api_url}")
    except requests.exceptions.ConnectionError:
        raise LookupError(f"PyPI API Connection error: {pip_api_url}")
    else:
        if response.status_code == 200:
            return response.json()
        raise ValueError(f"Could not find pip dependency using the PyPI API: `{dep}`")


def get_biocontainer_tag(package, version):
    """
    Given a bioconda package and version, looks for Docker and Singularity containers
    using the biocontaineres API, e.g.:
    https://api.biocontainers.pro/ga4gh/trs/v2/tools/{tool}/versions/{tool}-{version}
    Returns the most recent container versions by default.
    Args:
        package (str): A bioconda package name.
        version (str): Version of the bioconda package
    Raises:
        A LookupError, if the connection fails or times out or gives an unexpected status code
        A ValueError, if the package name can not be found (404)
    """

    biocontainers_api_url = f"https://api.biocontainers.pro/ga4gh/trs/v2/tools/{package}/versions/{package}-{version}"

    def get_tag_date(tag_date):
        """
        Format a date given by the biocontainers API
        Given format: '2021-03-25T08:53:00Z'
        """
        return datetime.datetime.strptime(tag_date, "%Y-%m-%dT%H:%M:%SZ")

    try:
        response = requests.get(biocontainers_api_url)
    except requests.exceptions.ConnectionError:
        raise LookupError("Could not connect to biocontainers.pro API")
    else:
        if response.status_code == 200:
            try:
                images = response.json()["images"]
                singularity_image = None
                docker_image = None
                all_docker = {}
                all_singularity = {}
                for img in images:
                    # Get all Docker and Singularity images
                    if img["image_type"] == "Docker":
                        # Obtain version and build
                        match = re.search(r"(?::)+([A-Za-z\d\-_.]+)", img["image_name"])
                        if match is not None:
                            all_docker[match.group(1)] = {
                                "date": get_tag_date(img["updated"]),
                                "image": img,
                            }
                    elif img["image_type"] == "Singularity":
                        # Obtain version and build
                        match = re.search(r"(?::)+([A-Za-z\d\-_.]+)", img["image_name"])
                        if match is not None:
                            all_singularity[match.group(1)] = {
                                "date": get_tag_date(img["updated"]),
                                "image": img,
                            }
                # Obtain common builds from Docker and Singularity images
                common_keys = list(all_docker.keys() & all_singularity.keys())
                current_date = None
                docker_image_name = docker_image["image_name"].lstrip("quay.io/") if docker_image is not None else None
                for k in common_keys:
                    # Get the most recent common image
                    date = max(all_docker[k]["date"], all_docker[k]["date"])
                    if docker_image is None or current_date < date:
                        docker_image = all_docker[k]["image"]
                        singularity_image = all_singularity[k]["image"]
                        current_date = date
                        docker_image_name = docker_image["image_name"].lstrip("quay.io/")
                if singularity_image is None:
                    raise LookupError(f"Could not find singularity container for {package}")
                return docker_image_name, singularity_image["image_name"]
            except TypeError:
                raise LookupError(f"Could not find docker or singularity container for {package}")
        elif response.status_code != 404:
            raise LookupError(f"Unexpected response code `{response.status_code}` for {biocontainers_api_url}")
        elif response.status_code == 404:
            raise ValueError(f"Could not find `{package}` on api.biocontainers.pro")


def custom_yaml_dumper():
    """Overwrite default PyYAML output to make Prettier YAML linting happy"""

    class CustomDumper(yaml.Dumper):
        def represent_dict_preserve_order(self, data):
            """Add custom dumper class to prevent overwriting the global state
            This prevents yaml from changing the output order

            See https://stackoverflow.com/a/52621703/1497385
            """
            return self.represent_dict(data.items())

        def increase_indent(self, flow=False, indentless=False):
            """Indent YAML lists so that YAML validates with Prettier

            See https://github.com/yaml/pyyaml/issues/234#issuecomment-765894586
            """
            return super().increase_indent(flow=flow, indentless=False)

        # HACK: insert blank lines between top-level objects
        # inspired by https://stackoverflow.com/a/44284819/3786245
        # and https://github.com/yaml/pyyaml/issues/127
        def write_line_break(self, data=None):
            super().write_line_break(data)

            if len(self.indents) == 1:
                super().write_line_break()

    CustomDumper.add_representer(dict, CustomDumper.represent_dict_preserve_order)
    return CustomDumper


def is_file_binary(path):
    """Check file path to see if it is a binary file"""
    binary_ftypes = ["image", "application/java-archive", "application/x-java-archive"]
    binary_extensions = [".jpeg", ".jpg", ".png", ".zip", ".gz", ".jar", ".tar"]

    # Check common file extensions
    _, file_extension = os.path.splitext(path)
    if file_extension in binary_extensions:
        return True

    # Try to detect binary files
    (ftype, encoding) = mimetypes.guess_type(path, strict=False)
    if encoding is not None or (ftype is not None and any(ftype.startswith(ft) for ft in binary_ftypes)):
        return True


def prompt_remote_pipeline_name(wfs):
    """Prompt for the pipeline name with questionary

    Args:
        wfs: A nf_core.pipelines.list.Workflows() object, where get_remote_workflows() has been called.

    Returns:
        pipeline (str): GitHub repo - username/repo

    Raises:
        AssertionError, if pipeline cannot be found
    """

    pipeline = questionary.autocomplete(
        "Pipeline name:",
        choices=[wf.name for wf in wfs.remote_workflows],
        style=nfcore_question_style,
    ).unsafe_ask()

    # Check nf-core repos
    for wf in wfs.remote_workflows:
        if wf.full_name == pipeline or wf.name == pipeline:
            return wf.full_name

    # Non nf-core repo on GitHub
    if pipeline.count("/") == 1:
        try:
            gh_api.safe_get(f"https://api.github.com/repos/{pipeline}")
        except Exception:
            # No repo found - pass and raise error at the end
            pass
        else:
            return pipeline

    log.info("Available nf-core pipelines: '{}'".format("', '".join([w.name for w in wfs.remote_workflows])))
    raise AssertionError(f"Not able to find pipeline '{pipeline}'")


def prompt_pipeline_release_branch(
    wf_releases: List[Dict[str, Any]], wf_branches: Dict[str, Any], multiple: bool = False
) -> Tuple[Any, List[str]]:
    """Prompt for pipeline release / branch

    Args:
        wf_releases (array): Array of repo releases as returned by the GitHub API
        wf_branches (array): Array of repo branches, as returned by the GitHub API
        multiple (bool): Allow selection of multiple releases & branches (for Seqera Platform)

    Returns:
        choice (questionary.Choice or bool): Selected release / branch or False if no releases / branches available
    """
    # Prompt user for release tag, tag_set will contain all available.
    choices: List[questionary.Choice] = []
    tag_set: List[str] = []

    # Releases
    if len(wf_releases) > 0:
        for tag in map(lambda release: release.get("tag_name"), wf_releases):
            tag_display = [
                ("fg:ansiblue", f"{tag}  "),
                ("class:choice-default", "[release]"),
            ]
            choices.append(questionary.Choice(title=tag_display, value=tag))
            tag_set.append(str(tag))

    # Branches
    for branch in wf_branches.keys():
        branch_display = [
            ("fg:ansiyellow", f"{branch}  "),
            ("class:choice-default", "[branch]"),
        ]
        choices.append(questionary.Choice(title=branch_display, value=branch))
        tag_set.append(branch)

    if len(choices) == 0:
        return [], []

    if multiple:
        return (
            questionary.checkbox("Select release / branch:", choices=choices, style=nfcore_question_style).unsafe_ask(),
            tag_set,
        )

    else:
        return (
            questionary.select("Select release / branch:", choices=choices, style=nfcore_question_style).unsafe_ask(),
            tag_set,
        )


class SingularityCacheFilePathValidator(questionary.Validator):
    """
    Validator for file path specified as --singularity-cache-index argument in nf-core pipelines download
    """

    def validate(self, value):
        if len(value.text):
            if os.path.isfile(value.text):
                return True
            else:
                raise questionary.ValidationError(
                    message="Invalid remote cache index file",
                    cursor_position=len(value.text),
                )
        else:
            return True


def get_repo_releases_branches(pipeline, wfs):
    """Fetches details of a nf-core workflow to download.

    Args:
        pipeline (str): GitHub repo username/repo
        wfs: A nf_core.pipelines.list.Workflows() object, where get_remote_workflows() has been called.

    Returns:
        wf_releases, wf_branches (tuple): Array of releases, Array of branches

    Raises:
        LockupError, if the pipeline can not be found.
    """

    wf_releases = []
    wf_branches = {}

    # Repo is a nf-core pipeline
    for wf in wfs.remote_workflows:
        if wf.full_name == pipeline or wf.name == pipeline:
            # Set to full name just in case it didn't have the nf-core/ prefix
            pipeline = wf.full_name

            # Store releases and stop loop
            wf_releases = list(
                sorted(
                    wf.releases,
                    key=lambda k: k.get("published_at_timestamp", 0),
                    reverse=True,
                )
            )
            break

    # Arbitrary GitHub repo
    else:
        if pipeline.count("/") == 1:
            # Looks like a GitHub address - try working with this repo
            log.debug(
                f"Pipeline '{pipeline}' not in nf-core, but looks like a GitHub address - fetching releases from API"
            )

            # Get releases from GitHub API
            rel_r = gh_api.safe_get(f"https://api.github.com/repos/{pipeline}/releases")

            # Check that this repo existed
            try:
                if rel_r.json().get("message") == "Not Found":
                    raise AssertionError(f"Not able to find pipeline '{pipeline}'")
            except AttributeError:
                # Success! We have a list, which doesn't work with .get() which is looking for a dict key
                wf_releases = list(
                    sorted(
                        rel_r.json(),
                        key=lambda k: k.get("published_at_timestamp", 0),
                        reverse=True,
                    )
                )

                # Get release tag commit hashes
                if len(wf_releases) > 0:
                    # Get commit hash information for each release
                    tags_r = gh_api.safe_get(f"https://api.github.com/repos/{pipeline}/tags")
                    for tag in tags_r.json():
                        for release in wf_releases:
                            if tag["name"] == release["tag_name"]:
                                release["tag_sha"] = tag["commit"]["sha"]

        else:
            log.info("Available nf-core pipelines: '{}'".format("', '".join([w.name for w in wfs.remote_workflows])))
            raise AssertionError(f"Not able to find pipeline '{pipeline}'")

    # Get branch information from github api - should be no need to check if the repo exists again
    branch_response = gh_api.safe_get(f"https://api.github.com/repos/{pipeline}/branches?per_page=100")
    for branch in branch_response.json():
        if (
            branch["name"] != "TEMPLATE"
            and branch["name"] != "initial_commit"
            and not branch["name"].startswith("nf-core-template-merge")
        ):
            wf_branches[branch["name"]] = branch["commit"]["sha"]

    # Return pipeline again in case we added the nf-core/ prefix
    return pipeline, wf_releases, wf_branches


CONFIG_PATHS = [".nf-core.yml", ".nf-core.yaml"]
DEPRECATED_CONFIG_PATHS = [".nf-core-lint.yml", ".nf-core-lint.yaml"]


class NFCoreTemplateConfig(BaseModel):
    """Template configuration schema"""

    org: Optional[str] = None
    """ Organisation name """
    name: Optional[str] = None
    """ Pipeline name """
    description: Optional[str] = None
    """ Pipeline description """
    author: Optional[str] = None
    """ Pipeline author """
    version: Optional[str] = None
    """ Pipeline version """
    force: Optional[bool] = True
    """ Force overwrite of existing files """
    outdir: Optional[Union[str, Path]] = None
    """ Output directory """
    skip_features: Optional[list] = None
    """ Skip features. See https://nf-co.re/docs/nf-core-tools/pipelines/create for a list of features. """
    is_nfcore: Optional[bool] = None
    """ Whether the pipeline is an nf-core pipeline. """

    # convert outdir to str
    @field_validator("outdir")
    @classmethod
    def outdir_to_str(cls, v: Optional[Union[str, Path]]) -> Optional[str]:
        if v is not None:
            v = str(v)
        return v

    def __getitem__(self, item: str) -> Any:
        if self is None:
            return None
        return getattr(self, item)

    def get(self, item: str, default: Any = None) -> Any:
        return getattr(self, item, default)


class NFCoreYamlLintConfig(BaseModel):
    """
    schema for linting config in `.nf-core.yml` should cover:

    .. code-block:: yaml
        files_unchanged:
            - .github/workflows/branch.yml
        modules_config: False
        modules_config:
                - fastqc
        # merge_markers: False
        merge_markers:
                - docs/my_pdf.pdf
        nextflow_config: False
        nextflow_config:
            - manifest.name
            - config_defaults:
                - params.annotation_db
                - params.multiqc_comment_headers
                - params.custom_table_headers
        # multiqc_config: False
        multiqc_config:
            - report_section_order
            - report_comment
        files_exist:
            - .github/CONTRIBUTING.md
            - CITATIONS.md
        template_strings: False
        template_strings:
                - docs/my_pdf.pdf
        nfcore_components: False
    """

    files_unchanged: Optional[Union[bool, List[str]]] = None
    """ List of files that should not be changed """
    modules_config: Optional[Optional[Union[bool, List[str]]]] = None
    """ List of modules that should not be changed """
    merge_markers: Optional[Optional[Union[bool, List[str]]]] = None
    """ List of files that should not contain merge markers """
    nextflow_config: Optional[Optional[Union[bool, List[Union[str, Dict[str, List[str]]]]]]] = None
    """ List of Nextflow config files that should not be changed """
    multiqc_config: Optional[Union[bool, List[str]]] = None
    """ List of MultiQC config options that be changed """
    files_exist: Optional[Union[bool, List[str]]] = None
    """ List of files that can not exist """
    template_strings: Optional[Optional[Union[bool, List[str]]]] = None
    """ List of files that can contain template strings """
    readme: Optional[Union[bool, List[str]]] = None
    """ Lint the README.md file """
    nfcore_components: Optional[bool] = None
    """ Lint all required files to use nf-core modules and subworkflows """
    actions_ci: Optional[bool] = None
    """ Lint all required files to use GitHub Actions CI """
    actions_awstest: Optional[bool] = None
    """ Lint all required files to run tests on AWS """
    actions_awsfulltest: Optional[bool] = None
    """ Lint all required files to run full tests on AWS """
    pipeline_todos: Optional[bool] = None
    """ Lint for TODOs statements"""
    plugin_includes: Optional[bool] = None
    """ Lint for nextflow plugin """
    pipeline_name_conventions: Optional[bool] = None
    """ Lint for pipeline name conventions """
    schema_lint: Optional[bool] = None
    """ Lint nextflow_schema.json file"""
    schema_params: Optional[bool] = None
    """ Lint schema for all params """
    system_exit: Optional[bool] = None
    """ Lint for System.exit calls in groovy/nextflow code """
    schema_description: Optional[bool] = None
    """ Check that every parameter in the schema has a description. """
    actions_schema_validation: Optional[bool] = None
    """ Lint GitHub Action workflow files with schema"""
    modules_json: Optional[bool] = None
    """ Lint modules.json file """
    modules_structure: Optional[bool] = None
    """ Lint modules structure """
    base_config: Optional[bool] = None
    """ Lint base.config file """
    nfcore_yml: Optional[bool] = None
    """ Lint nf-core.yml """
    version_consistency: Optional[bool] = None
    """ Lint for version consistency """
    included_configs: Optional[bool] = None
    """ Lint for included configs """

    def __getitem__(self, item: str) -> Any:
        return getattr(self, item)

    def get(self, item: str, default: Any = None) -> Any:
        if getattr(self, item, default) is None:
            return default
        return getattr(self, item, default)

    def __setitem__(self, item: str, value: Any) -> None:
        setattr(self, item, value)


class NFCoreYamlConfig(BaseModel):
    """.nf-core.yml configuration file schema"""

    repository_type: Optional[Literal["pipeline", "modules"]] = None
    """ Type of repository """
    nf_core_version: Optional[str] = None
    """ Version of nf-core/tools used to create/update the pipeline """
    org_path: Optional[str] = None
    """ Path to the organisation's modules repository (used for modules repo_type only) """
    lint: Optional[NFCoreYamlLintConfig] = None
    """ Pipeline linting configuration, see https://nf-co.re/docs/nf-core-tools/pipelines/lint#linting-config for examples and documentation """
    template: Optional[NFCoreTemplateConfig] = None
    """ Pipeline template configuration """
    bump_version: Optional[Dict[str, bool]] = None
    """ Disable bumping of the version for a module/subworkflow (when repository_type is modules). See https://nf-co.re/docs/nf-core-tools/modules/bump-versions for more information. """
    update: Optional[Dict[str, Union[str, bool, Dict[str, Union[str, Dict[str, Union[str, bool]]]]]]] = None
    """ Disable updating specific modules/subworkflows (when repository_type is pipeline). See https://nf-co.re/docs/nf-core-tools/modules/update for more information. """

    def __getitem__(self, item: str) -> Any:
        return getattr(self, item)

    def get(self, item: str, default: Any = None) -> Any:
        return getattr(self, item, default)

    def __setitem__(self, item: str, value: Any) -> None:
        setattr(self, item, value)

    def model_dump(self, **kwargs) -> Dict[str, Any]:
        # Get the initial data
        config = super().model_dump(**kwargs)

        if self.repository_type == "modules":
            # Fields to exclude for modules
            fields_to_exclude = ["template", "update"]
        else:  # pipeline
            # Fields to exclude for pipeline
            fields_to_exclude = ["bump_version", "org_path"]

        # Remove the fields based on repository_type
        for field in fields_to_exclude:
            config.pop(field, None)

        return config


def load_tools_config(directory: Union[str, Path] = ".") -> Tuple[Optional[Path], Optional[NFCoreYamlConfig]]:
    """
    Parse the nf-core.yml configuration file

    Look for a file called either `.nf-core.yml` or `.nf-core.yaml`

    Also looks for the deprecated file `.nf-core-lint.yml/yaml` and issues
    a warning that this file will be deprecated in the future

    Returns the loaded config dict or False, if the file couldn't be loaded
    """
    tools_config = {}

    config_fn = get_first_available_path(directory, CONFIG_PATHS)
    if config_fn is None:
        depr_path = get_first_available_path(directory, DEPRECATED_CONFIG_PATHS)
        if depr_path:
            raise UserWarning(
                f"Deprecated `{depr_path.name}` file found! Please rename the file to `{CONFIG_PATHS[0]}`."
            )
        else:
            log.debug(f"Could not find a config file in the directory '{directory}'")
            return Path(directory, CONFIG_PATHS[0]), None
    if not Path(config_fn).is_file():
        raise FileNotFoundError(f"No `.nf-core.yml` file found in the directory '{directory}'")
    with open(config_fn) as fh:
        tools_config = yaml.safe_load(fh)

    # If the file is empty
    if tools_config is None:
        raise AssertionError(f"Config file '{config_fn}' is empty")
    # Check for required fields
    try:
        nf_core_yaml_config = NFCoreYamlConfig(**tools_config)
    except ValidationError as e:
        error_message = f"Config file '{config_fn}' is invalid"
        for error in e.errors():
            error_message += f"\n{error['loc'][0]}: {error['msg']}\ninput: {error['input']}"
        raise AssertionError(error_message)

    wf_config = fetch_wf_config(Path(directory))
    if nf_core_yaml_config["repository_type"] == "pipeline" and wf_config:
        # Retrieve information if template from config file is empty
        template = tools_config.get("template")
        config_template_keys = template.keys() if template is not None else []
        # Get author names from contributors first, then fallback to author
        if "manifest.contributors" in wf_config:
            contributors = wf_config["manifest.contributors"]
            names = re.findall(r"name:'([^']+)'", contributors)
            author_names = ", ".join(names)
        else:
            author_names = wf_config["manifest.author"].strip("'\"")
        if nf_core_yaml_config.template is None:
            # The .nf-core.yml file did not contain template information
            nf_core_yaml_config.template = NFCoreTemplateConfig(
                org="nf-core",
                name=wf_config["manifest.name"].strip("'\"").split("/")[-1],
                description=wf_config["manifest.description"].strip("'\""),
                author=author_names,
                version=wf_config["manifest.version"].strip("'\""),
                outdir=str(directory),
                is_nfcore=True,
            )
        elif "prefix" in config_template_keys or "skip" in config_template_keys:
            # The .nf-core.yml file contained the old prefix or skip keys
            nf_core_yaml_config.template = NFCoreTemplateConfig(
                org=tools_config["template"].get("prefix", tools_config["template"].get("org", "nf-core")),
                name=tools_config["template"].get("name", wf_config["manifest.name"].strip("'\"").split("/")[-1]),
                description=tools_config["template"].get("description", wf_config["manifest.description"].strip("'\"")),
                author=tools_config["template"].get("author", author_names),
                version=tools_config["template"].get("version", wf_config["manifest.version"].strip("'\"")),
                outdir=tools_config["template"].get("outdir", str(directory)),
                skip_features=tools_config["template"].get("skip", tools_config["template"].get("skip_features")),
                is_nfcore=tools_config["template"].get("prefix", tools_config["template"].get("org")) == "nf-core",
            )

    log.debug("Using config file: %s", config_fn)
    return config_fn, nf_core_yaml_config


def determine_base_dir(directory: Union[Path, str] = ".") -> Path:
    base_dir = start_dir = Path(directory).absolute()
    # Only iterate up the tree if the start dir doesn't have a config
    while not get_first_available_path(base_dir, CONFIG_PATHS) and base_dir != base_dir.parent:
        base_dir = base_dir.parent
        config_fn = get_first_available_path(base_dir, CONFIG_PATHS)
        if config_fn:
            break
    return Path(directory) if (base_dir == start_dir or str(base_dir) == base_dir.root) else base_dir


def get_first_available_path(directory: Union[Path, str], paths: List[str]) -> Union[Path, None]:
    for p in paths:
        if Path(directory, p).is_file():
            return Path(directory, p)
    return None


def sort_dictionary(d: Dict) -> Dict:
    """Sorts a nested dictionary recursively"""
    result = {}
    for k, v in sorted(d.items()):
        if isinstance(v, dict):
            result[k] = sort_dictionary(v)
        else:
            result[k] = v
    return result


def plural_s(list_or_int):
    """Return an s if the input is not one or has not the length of one."""
    length = list_or_int if isinstance(list_or_int, int) else len(list_or_int)
    return "s" * (length != 1)


def plural_y(list_or_int):
    """Return 'ies' if the input is not one or has not the length of one, else 'y'."""
    length = list_or_int if isinstance(list_or_int, int) else len(list_or_int)
    return "ies" if length != 1 else "y"


def plural_es(list_or_int):
    """Return a 'es' if the input is not one or has not the length of one."""
    length = list_or_int if isinstance(list_or_int, int) else len(list_or_int)
    return "es" * (length != 1)


# From Stack Overflow: https://stackoverflow.com/a/14693789/713980
# Placed at top level as to only compile it once
ANSI_ESCAPE_RE = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")


def strip_ansi_codes(string, replace_with=""):
    """Strip ANSI colouring codes from a string to return plain text.

    From Stack Overflow: https://stackoverflow.com/a/14693789/713980
    """
    return ANSI_ESCAPE_RE.sub(replace_with, string)


def is_relative_to(path1, path2):
    """
    Checks if a path is relative to another.

    Should mimic Path.is_relative_to which not available in Python < 3.9

    path1 (Path | str): The path that could be a subpath
    path2 (Path | str): The path the could be the superpath
    """
    return str(path1).startswith(str(path2) + os.sep)


def file_md5(fname):
    """Calculates the md5sum for a file on the disk.

    Args:
        fname (str): Path to a local file.
    """

    # Calculate the md5 for the file on disk
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(io.DEFAULT_BUFFER_SIZE), b""):
            hash_md5.update(chunk)

    return hash_md5.hexdigest()


def validate_file_md5(file_name, expected_md5hex):
    """Validates the md5 checksum of a file on disk.

    Args:
        file_name (str): Path to a local file.
        expected (str): The expected md5sum.

    Raises:
        IOError, if the md5sum does not match the remote sum.
    """
    log.debug(f"Validating image hash: {file_name}")

    # Make sure the expected md5 sum is a hexdigest
    try:
        int(expected_md5hex, 16)
    except ValueError as ex:
        raise ValueError(f"The supplied md5 sum must be a hexdigest but it is {expected_md5hex}") from ex

    file_md5hex = file_md5(file_name)

    if file_md5hex.upper() == expected_md5hex.upper():
        log.debug(f"md5 sum of image matches expected: {expected_md5hex}")
    else:
        raise OSError(f"{file_name} md5 does not match remote: {expected_md5hex} - {file_md5hex}")

    return True


def nested_setitem(d, keys, value):
    """Sets the value in a nested dict using a list of keys to traverse

    Args:
        d (dict): the nested dictionary to traverse
        keys (list[Any]): A list of keys to iteratively traverse
        value (Any): The value to be set for the last key in the chain
    """
    current = d
    for k in keys[:-1]:
        current = current[k]
    current[keys[-1]] = value


def nested_delitem(d, keys):
    """Deletes a key from a nested dictionary

    Args:
        d (dict): the nested dictionary to traverse
        keys (list[Any]): A list of keys to iteratively traverse, deleting the final one
    """
    current = d
    for k in keys[:-1]:
        current = current[k]
    del current[keys[-1]]


@contextmanager
def set_wd(path: Path) -> Generator[None, None, None]:
    """Sets the working directory for this context.

    Arguments
    ---------

    path : Path
        Path to the working directory to be used inside this context.
    """
    start_wd = Path().absolute()
    os.chdir(Path(path).resolve())
    try:
        yield
    finally:
        os.chdir(start_wd)


def get_wf_files(wf_path: Path):
    """Return a list of all files in a directory (ignores .gitigore files)"""

    wf_files = []

    with open(Path(wf_path, ".gitignore")) as f:
        lines = f.read().splitlines()
    ignore = [line for line in lines if line and not line.startswith("#")]

    for path in Path(wf_path).rglob("*"):
        if any(fnmatch.fnmatch(str(path), pattern) for pattern in ignore):
            continue
        if path.is_file():
            wf_files.append(str(path))

    return wf_files



from rich.console import Console

import nf_core.subworkflows

from ..test_subworkflows import TestSubworkflows
from ..utils import GITLAB_SUBWORKFLOWS_BRANCH, GITLAB_URL


class TestSubworkflowsInfo(TestSubworkflows):
    def test_subworkflows_info_remote(self):
        """Test getting info about a remote subworkflow"""
        mods_info = nf_core.subworkflows.SubworkflowInfo(self.pipeline_dir, "bam_sort_stats_samtools")
        mods_info_output = mods_info.get_component_info()
        console = Console(record=True)
        console.print(mods_info_output)
        output = console.export_text()

        assert "Subworkflow: bam_sort_stats_samtools" in output
        assert "Inputs" in output
        assert "Outputs" in output

    def test_subworkflows_info_remote_gitlab(self):
        """Test getting info about a subworkflow in the remote gitlab repo"""
        mods_info = nf_core.subworkflows.SubworkflowInfo(
            self.pipeline_dir, "bam_sort_stats_samtools", remote_url=GITLAB_URL, branch=GITLAB_SUBWORKFLOWS_BRANCH
        )
        mods_info_output = mods_info.get_component_info()
        console = Console(record=True)
        console.print(mods_info_output)
        output = console.export_text()

        assert "Subworkflow: bam_sort_stats_samtools" in output
        assert "Inputs" in output
        assert "Outputs" in output
        assert "--git-remote" in output

    def test_subworkflows_info_local(self):
        """Test getting info about a locally installed subworkflow"""
        self.subworkflow_install.install("bam_sort_stats_samtools")
        mods_info = nf_core.subworkflows.SubworkflowInfo(self.pipeline_dir, "bam_sort_stats_samtools")
        mods_info.local = True
        mods_info_output = mods_info.get_component_info()
        console = Console(record=True)
        console.print(mods_info_output)
        output = console.export_text()

        assert "Subworkflow: bam_sort_stats_samtools" in output
        assert "Inputs" in output
        assert "Outputs" in output

    def test_subworkflows_info_in_modules_repo(self):
        """Test getting info about a locally subworkflow in the modules repo"""
        self.subworkflow_install.install("bam_sort_stats_samtools")
        mods_info = nf_core.subworkflows.SubworkflowInfo(self.nfcore_modules, "bam_sort_stats_samtools")
        mods_info.local = True
        mods_info_output = mods_info.get_component_info()
        console = Console(record=True)
        console.print(mods_info_output)
        output = console.export_text()

        assert "Subworkflow: bam_sort_stats_samtools" in output
        assert "Inputs" in output
        assert "Outputs" in output



"""Tests covering the pipeline schema code."""

import json
import os
import shutil
import tempfile
import unittest
from pathlib import Path
from unittest import mock

import pytest
import requests
import yaml

import nf_core.pipelines.create.create
import nf_core.pipelines.schema

from ..utils import with_temporary_file, with_temporary_folder


class TestSchema(unittest.TestCase):
    """Class for schema tests"""

    def setUp(self):
        """Create a new PipelineSchema object"""
        self.schema_obj = nf_core.pipelines.schema.PipelineSchema()
        self.schema_obj.schema_draft = "https://json-schema.org/draft/2020-12/schema"
        self.schema_obj.defs_notation = "$defs"
        self.schema_obj.validation_plugin = "nf-schema"
        self.root_repo_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

        # Create a test pipeline in temp directory
        self.tmp_dir = tempfile.mkdtemp()
        self.template_dir = os.path.join(self.tmp_dir, "wf")
        create_obj = nf_core.pipelines.create.create.PipelineCreate(
            "testpipeline", "a description", "Me", outdir=self.template_dir, no_git=True
        )
        create_obj.init_pipeline()

        self.template_schema = os.path.join(self.template_dir, "nextflow_schema.json")

    def tearDown(self):
        if os.path.exists(self.tmp_dir):
            shutil.rmtree(self.tmp_dir)

    def test_load_lint_schema(self):
        """Check linting with the pipeline template directory"""
        self.schema_obj.get_schema_path(self.template_dir)
        self.schema_obj.load_lint_schema()

    def test_load_lint_schema_nofile(self):
        """Check that linting raises properly if a non-existent file is given"""
        with pytest.raises(RuntimeError):
            self.schema_obj.get_schema_path("fake_file")

    def test_load_lint_schema_notjson(self):
        """Check that linting raises properly if a non-JSON file is given"""
        self.schema_obj.get_schema_path(os.path.join(self.template_dir, "nextflow.config"))
        with pytest.raises(AssertionError):
            self.schema_obj.load_lint_schema()

    @with_temporary_file
    def test_load_lint_schema_noparams(self, tmp_file):
        """
        Check that linting raises properly if a JSON file is given without any params
        """
        # write schema to a temporary file
        with open(tmp_file.name, "w") as fh:
            json.dump({"type": "fubar"}, fh)
        self.schema_obj.get_schema_path(tmp_file.name)
        with pytest.raises(AssertionError):
            self.schema_obj.load_lint_schema()

    def test_get_schema_path_dir(self):
        """Get schema file from directory"""
        self.schema_obj.get_schema_path(self.template_dir)

    def test_get_schema_path_path(self):
        """Get schema file from a path"""
        self.schema_obj.get_schema_path(self.template_schema)

    def test_get_schema_path_path_notexist(self):
        """Get schema file from a path"""
        with pytest.raises(AssertionError):
            self.schema_obj.get_schema_path("fubar", local_only=True)

    def test_get_schema_path_name(self):
        """Get schema file from the name of a remote pipeline"""
        self.schema_obj.get_schema_path("atacseq")

    def test_get_schema_path_name_notexist(self):
        """
        Get schema file from the name of a remote pipeline
        that doesn't have a schema file
        """
        with pytest.raises(AssertionError):
            self.schema_obj.get_schema_path("exoseq")

    def test_load_schema(self):
        """Try to load a schema from a file"""
        self.schema_obj.schema_filename = self.template_schema
        self.schema_obj.load_schema()

    def test_schema_docs(self):
        """Try to generate Markdown docs for a schema from a file"""
        self.schema_obj.schema_filename = self.template_schema
        self.schema_obj.load_schema()
        docs = self.schema_obj.print_documentation()
        assert self.schema_obj.schema["title"] in docs
        assert self.schema_obj.schema["description"] in docs
        for definition in self.schema_obj.schema.get("$defs", {}).values():
            assert definition["title"] in docs
            assert definition["description"] in docs

    @with_temporary_file
    def test_save_schema(self, tmp_file):
        """Try to save a schema"""
        # Load the template schema
        self.schema_obj.schema_filename = self.template_schema
        self.schema_obj.load_schema()

        # Make a temporary file to write schema to
        self.schema_obj.schema_filename = tmp_file.name
        self.schema_obj.save_schema()

    @with_temporary_file
    def test_load_input_params_json(self, tmp_file):
        """Try to load a JSON file with params for a pipeline run"""
        # write schema to a temporary file
        with open(tmp_file.name, "w") as fh:
            json.dump({"input": "fubar"}, fh)
        self.schema_obj.load_input_params(tmp_file.name)

    @with_temporary_file
    def test_load_input_params_yaml(self, tmp_file):
        """Try to load a YAML file with params for a pipeline run"""
        # write schema to a temporary file
        with open(tmp_file.name, "w") as fh:
            yaml.dump({"input": "fubar"}, fh)
        self.schema_obj.load_input_params(tmp_file.name)

    def test_load_input_params_invalid(self):
        """Check failure when a non-existent file params file is loaded"""
        with pytest.raises(AssertionError):
            self.schema_obj.load_input_params("fubar")

    def test_validate_params_pass(self):
        """Try validating a set of parameters against a schema"""
        # Load the template schema
        self.schema_obj.schema_filename = self.template_schema
        self.schema_obj.load_schema()
        self.schema_obj.input_params = {"input": "fubar.csv", "outdir": "results/"}
        assert self.schema_obj.validate_params()

    def test_validate_params_fail(self):
        """Check that False is returned if params don't validate against a schema"""
        # Load the template schema
        self.schema_obj.schema_filename = self.template_schema
        self.schema_obj.load_schema()
        self.schema_obj.input_params = {"fubar": "input"}
        assert not self.schema_obj.validate_params()

    def test_validate_schema_pass(self):
        """Check that the schema validation passes"""
        # Load the template schema
        self.schema_obj.schema_filename = self.template_schema
        self.schema_obj.load_schema()
        self.schema_obj.validate_schema(self.schema_obj.schema)

    def test_validate_schema_fail_noparams(self):
        """Check that the schema validation fails when no params described"""
        self.schema_obj.schema = {"type": "invalidthing"}
        with pytest.raises(AssertionError):
            self.schema_obj.validate_schema(self.schema_obj.schema)

    def test_validate_schema_fail_duplicate_ids(self):
        """
        Check that the schema validation fails when we have duplicate IDs in definition subschema
        """
        self.schema_obj.schema = {
            "$schema": "https://json-schema.org/draft/2020-12/schema",
            "$defs": {"groupOne": {"properties": {"foo": {}}}, "groupTwo": {"properties": {"foo": {}}}},
            "allOf": [{"$ref": "#/$defs/groupOne"}, {"$ref": "#/$defs/groupTwo"}],
        }
        with pytest.raises(AssertionError) as exc_info:
            self.schema_obj.validate_schema(self.schema_obj.schema)
        assert exc_info.value.args[0] == "Duplicate parameter found in schema `$defs`: `foo`"

    def test_validate_schema_fail_missing_def(self):
        """
        Check that the schema validation fails when we a definition in allOf is not in $defs
        """
        self.schema_obj.schema = {
            "$schema": "https://json-schema.org/draft/2020-12/schema",
            "$defs": {"groupOne": {"properties": {"foo": {}}}, "groupTwo": {"properties": {"bar": {}}}},
            "allOf": [{"$ref": "#/$defs/groupOne"}],
        }
        with pytest.raises(AssertionError) as exc_info:
            self.schema_obj.validate_schema(self.schema_obj.schema)
        assert exc_info.value.args[0] == "Definition subschema `#/$defs/groupTwo` not included in schema `allOf`"

    def test_validate_schema_fail_unexpected_allof(self):
        """
        Check that the schema validation fails when we an unrecognised definition is in allOf
        """
        self.schema_obj.schema = {
            "$schema": "https://json-schema.org/draft/2020-12/schema",
            "$defs": {"groupOne": {"properties": {"foo": {}}}, "groupTwo": {"properties": {"bar": {}}}},
            "allOf": [
                {"$ref": "#/$defs/groupOne"},
                {"$ref": "#/$defs/groupTwo"},
                {"$ref": "#/$defs/groupThree"},
            ],
        }
        with pytest.raises(AssertionError) as exc_info:
            self.schema_obj.validate_schema(self.schema_obj.schema)
        assert exc_info.value.args[0] == "Subschema `groupThree` found in `allOf` but not `$defs`"

    def test_make_skeleton_schema(self):
        """Test making a new schema skeleton"""
        self.schema_obj.schema_filename = self.template_schema
        self.schema_obj.pipeline_manifest["name"] = "nf-core/test"
        self.schema_obj.pipeline_manifest["description"] = "Test pipeline"
        self.schema_obj.make_skeleton_schema()
        self.schema_obj.validate_schema(self.schema_obj.schema)
        assert self.schema_obj.schema["title"] == "nf-core/test pipeline parameters"

    def test_make_skeleton_schema_absent_name(self):
        """Test making a new schema skeleton"""
        self.schema_obj.schema_filename = self.template_schema
        self.schema_obj.pipeline_manifest["description"] = "Test pipeline"
        self.schema_obj.make_skeleton_schema()
        self.schema_obj.validate_schema(self.schema_obj.schema)
        assert self.schema_obj.schema["title"] == "wf pipeline parameters"

    def test_get_wf_params(self):
        """Test getting the workflow parameters from a pipeline"""
        self.schema_obj.schema_filename = self.template_schema
        self.schema_obj.get_wf_params()

    def test_prompt_remove_schema_notfound_config_returntrue(self):
        """Remove unrecognised params from the schema"""
        self.schema_obj.pipeline_params = {"foo": "bar"}
        self.schema_obj.no_prompts = True
        assert self.schema_obj.prompt_remove_schema_notfound_config("baz")

    def test_prompt_remove_schema_notfound_config_returnfalse(self):
        """Do not remove unrecognised params from the schema"""
        self.schema_obj.pipeline_params = {"foo": "bar"}
        self.schema_obj.no_prompts = True
        assert not self.schema_obj.prompt_remove_schema_notfound_config("foo")

    def test_remove_schema_notfound_configs(self):
        """Remove unrecognised params from the schema"""
        self.schema_obj.schema = {
            "properties": {"foo": {"type": "string"}, "bar": {"type": "string"}},
            "required": ["foo"],
        }
        self.schema_obj.pipeline_params = {"bar": True}
        self.schema_obj.no_prompts = True
        params_removed = self.schema_obj.remove_schema_notfound_configs()
        assert len(self.schema_obj.schema["properties"]) == 1
        assert "required" not in self.schema_obj.schema
        assert len(params_removed) == 1
        assert "foo" in params_removed

    def test_remove_schema_notfound_configs_childschema(self):
        """
        Remove unrecognised params from the schema,
        even when they're in a group
        """
        self.schema_obj.schema = {
            "$defs": {
                "subSchemaId": {
                    "properties": {"foo": {"type": "string"}, "bar": {"type": "string"}},
                    "required": ["foo"],
                }
            }
        }
        self.schema_obj.pipeline_params = {"bar": True}
        self.schema_obj.no_prompts = True
        params_removed = self.schema_obj.remove_schema_notfound_configs()
        assert len(self.schema_obj.schema["$defs"]["subSchemaId"]["properties"]) == 1
        assert "required" not in self.schema_obj.schema["$defs"]["subSchemaId"]
        assert len(params_removed) == 1
        assert "foo" in params_removed

    def test_add_schema_found_configs(self):
        """Try adding a new parameter to the schema from the config"""
        self.schema_obj.pipeline_params = {"foo": "bar"}
        self.schema_obj.schema = {"properties": {}}
        self.schema_obj.no_prompts = True
        params_added = self.schema_obj.add_schema_found_configs()
        assert len(self.schema_obj.schema["properties"]) == 1
        assert len(params_added) == 1
        assert "foo" in params_added

    def test_build_schema_param_str(self):
        """Build a new schema param from a config value (string)"""
        param = self.schema_obj.build_schema_param("foo")
        assert param == {"type": "string", "default": "foo"}

    def test_build_schema_param_bool(self):
        """Build a new schema param from a config value (bool)"""
        param = self.schema_obj.build_schema_param("True")
        assert param == {"type": "boolean", "default": True}

    def test_build_schema_param_int(self):
        """Build a new schema param from a config value (int)"""
        param = self.schema_obj.build_schema_param("12")
        assert param == {"type": "integer", "default": 12}

    def test_build_schema_param_float(self):
        """Build a new schema param from a config value (float)"""
        param = self.schema_obj.build_schema_param("12.34")
        assert param == {"type": "number", "default": 12.34}

    def test_build_schema(self):
        """
        Build a new schema param from a pipeline
        Run code to ensure it doesn't crash. Individual functions tested separately.
        """
        self.schema_obj.build_schema(self.template_dir, True, False, None)

    @with_temporary_folder
    def test_build_schema_from_scratch(self, tmp_dir):
        """
        Build a new schema param from a pipeline with no existing file
        Run code to ensure it doesn't crash. Individual functions tested separately.

        Pretty much a copy of test_launch.py test_make_pipeline_schema
        """
        test_pipeline_dir = Path(tmp_dir, "wf")
        shutil.copytree(self.template_dir, test_pipeline_dir)
        Path(test_pipeline_dir, "nextflow_schema.json").unlink()

        self.schema_obj.build_schema(test_pipeline_dir, True, False, None)

    @mock.patch("requests.post")
    def test_launch_web_builder_timeout(self, mock_post):
        """Mock launching the web builder, but timeout on the request"""
        # Define the behaviour of the request get mock
        mock_post.side_effect = requests.exceptions.Timeout()
        with pytest.raises(AssertionError):
            self.schema_obj.launch_web_builder()

    @mock.patch("requests.post")
    def test_launch_web_builder_connection_error(self, mock_post):
        """Mock launching the web builder, but get a connection error"""
        # Define the behaviour of the request get mock
        mock_post.side_effect = requests.exceptions.ConnectionError()
        with pytest.raises(AssertionError):
            self.schema_obj.launch_web_builder()

    @mock.patch("requests.post")
    def test_get_web_builder_response_timeout(self, mock_post):
        """Mock checking for a web builder response, but timeout on the request"""
        # Define the behaviour of the request get mock
        mock_post.side_effect = requests.exceptions.Timeout()
        with pytest.raises(AssertionError):
            self.schema_obj.launch_web_builder()

    @mock.patch("requests.post")
    def test_get_web_builder_response_connection_error(self, mock_post):
        """Mock checking for a web builder response, but get a connection error"""
        # Define the behaviour of the request get mock
        mock_post.side_effect = requests.exceptions.ConnectionError()
        with pytest.raises(AssertionError):
            self.schema_obj.launch_web_builder()

    def mocked_requests_post(**kwargs):
        """Helper function to emulate POST requests responses from the web"""

        class MockResponse:
            def __init__(self, data, status_code):
                self.status_code = status_code
                self.content = json.dumps(data)

        if kwargs["url"] == "invalid_url":
            return MockResponse({}, 404)

        if kwargs["url"] == "valid_url_error":
            response_data = {"status": "error", "api_url": "foo", "web_url": "bar"}
            return MockResponse(response_data, 200)

        if kwargs["url"] == "valid_url_success":
            response_data = {"status": "recieved", "api_url": "https://nf-co.re", "web_url": "https://nf-co.re"}
            return MockResponse(response_data, 200)

    @mock.patch("requests.post", side_effect=mocked_requests_post)
    def test_launch_web_builder_404(self, mock_post):
        """Mock launching the web builder"""
        self.schema_obj.web_schema_build_url = "invalid_url"
        with pytest.raises(AssertionError) as exc_info:
            self.schema_obj.launch_web_builder()
        assert exc_info.value.args[0] == "Could not access remote API results: invalid_url (HTML 404 Error)"

    @mock.patch("requests.post", side_effect=mocked_requests_post)
    def test_launch_web_builder_invalid_status(self, mock_post):
        """Mock launching the web builder"""
        self.schema_obj.web_schema_build_url = "valid_url_error"
        with pytest.raises(AssertionError) as exc_info:
            self.schema_obj.launch_web_builder()
        assert exc_info.value.args[0].startswith("Pipeline schema builder response not recognised")

    @mock.patch("requests.post", side_effect=mocked_requests_post)
    @mock.patch("requests.get")
    @mock.patch("webbrowser.open")
    def test_launch_web_builder_success(self, mock_post, mock_get, mock_webbrowser):
        """Mock launching the web builder"""
        self.schema_obj.web_schema_build_url = "valid_url_success"
        # Assertion error comes from the get_web_builder_response() function
        with pytest.raises(AssertionError) as exc_info:
            self.schema_obj.launch_web_builder()
        assert exc_info.value.args[0].startswith("Could not access remote API results: https://nf-co.re")

    def mocked_requests_get(*args, **kwargs):
        """Helper function to emulate GET requests responses from the web"""

        class MockResponse:
            def __init__(self, data, status_code):
                self.status_code = status_code
                self.content = json.dumps(data)

        if args[0] == "invalid_url":
            return MockResponse({}, 404)

        if args[0] == "valid_url_error":
            response_data = {"status": "error", "message": "testing URL failure"}
            return MockResponse(response_data, 200)

        if args[0] == "valid_url_waiting":
            response_data = {"status": "waiting_for_user", "message": "testing URL waiting"}
            return MockResponse(response_data, 200)

        if args[0] == "valid_url_saved":
            response_data = {"status": "web_builder_edited", "message": "testing saved", "schema": {"foo": "bar"}}
            return MockResponse(response_data, 200)

    @mock.patch("requests.get", side_effect=mocked_requests_get)
    def test_get_web_builder_response_404(self, mock_post):
        """Mock launching the web builder"""
        self.schema_obj.web_schema_build_api_url = "invalid_url"
        with pytest.raises(AssertionError) as exc_info:
            self.schema_obj.get_web_builder_response()
        assert exc_info.value.args[0] == "Could not access remote API results: invalid_url (HTML 404 Error)"

    @mock.patch("requests.get", side_effect=mocked_requests_get)
    def test_get_web_builder_response_error(self, mock_post):
        """Mock launching the web builder"""
        self.schema_obj.web_schema_build_api_url = "valid_url_error"
        with pytest.raises(AssertionError) as exc_info:
            self.schema_obj.get_web_builder_response()
        assert exc_info.value.args[0] == "Got error from schema builder: 'testing URL failure'"

    @mock.patch("requests.get", side_effect=mocked_requests_get)
    def test_get_web_builder_response_waiting(self, mock_post):
        """Mock launching the web builder"""
        self.schema_obj.web_schema_build_api_url = "valid_url_waiting"
        assert self.schema_obj.get_web_builder_response() is False

    @mock.patch("requests.get", side_effect=mocked_requests_get)
    def test_get_web_builder_response_saved(self, mock_post):
        """Mock launching the web builder"""
        self.schema_obj.web_schema_build_api_url = "valid_url_saved"
        with pytest.raises(AssertionError) as exc_info:
            self.schema_obj.get_web_builder_response()
        assert exc_info.value.args[0].startswith("Response from schema builder did not pass validation")
        assert self.schema_obj.schema == {"foo": "bar"}



"""Create a YAML parameter file"""

import json
import logging
import textwrap
from pathlib import Path
from typing import Dict, List, Literal, Optional

import questionary

import nf_core.pipelines.list
import nf_core.utils
from nf_core.pipelines.schema import PipelineSchema

log = logging.getLogger(__name__)

INTRO = (
    "This is an example parameter file to pass to the `-params-file` option "
    "of nextflow run with the {pipeline_name} pipeline."
)

USAGE = "Uncomment lines with a single '#' if you want to pass the parameter " "to the pipeline."

H1_SEPERATOR = "## ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
H2_SEPERATOR = "## ----------------------------------------------------------------------------"

ModeLiteral = Literal["both", "start", "end", "none"]


def _print_wrapped(text, fill_char="-", mode="both", width=80, indent=0, drop_whitespace=True) -> str:
    """Helper function to format text for the params-file template.

    Args:
        text (str): Text to print
        fill_char (str, optional):
            Character to use for creating dividers. Defaults to '-'.
        mode (str, optional):
            Where to place dividers. Defaults to "both".
        width (int, optional):
            Maximum line-width of the output text. Defaults to 80.
        indent (int, optional):
            Number of spaces to indent the text. Defaults to 0.
        drop_whitespace (bool, optional):
            Whether to drop whitespace from the start and end of lines.
    """
    if len(fill_char) != 1:
        raise ValueError("fill_char must be a single character")

    prefix = "## "
    out = ""

    if mode in ("both", "start"):
        out += prefix.ljust(width, fill_char) + "\n"

    wrap_indent = f"{prefix}{' ' * indent}"

    textlines = textwrap.wrap(
        text,
        width=width - len(prefix),
        initial_indent=wrap_indent,
        subsequent_indent=wrap_indent,
        drop_whitespace=drop_whitespace,
    )

    for line in textlines:
        out += line + "\n"

    if mode in ("both", "end"):
        out += prefix.ljust(width, fill_char) + "\n"

    return out


class ParamsFileBuilder:
    """Class to hold config option to launch a pipeline.

    Args:
        pipeline (str, optional):
            Path to a local pipeline path or a remote pipeline.
        revision (str, optional):
            Revision of the pipeline to use.
    """

    def __init__(
        self,
        pipeline=None,
        revision=None,
    ) -> None:
        """Initialise the ParamFileBuilder class

        Args:
            pipeline (str, optional): Path to a local pipeline path or a remote pipeline.
            revision (str, optional): Revision of the pipeline to use.
        """
        self.pipeline = pipeline
        self.pipeline_revision = revision
        self.schema_obj: Optional[PipelineSchema] = None

        # Fetch remote workflows
        self.wfs = nf_core.pipelines.list.Workflows()
        self.wfs.get_remote_workflows()

    def get_pipeline(self) -> Optional[bool]:
        """
        Prompt the user for a pipeline name and get the schema
        """
        # Prompt for pipeline if not supplied
        if self.pipeline is None:
            launch_type = questionary.select(
                "Generate parameter file for local pipeline " "or remote GitHub pipeline?",
                choices=["Remote pipeline", "Local path"],
                style=nf_core.utils.nfcore_question_style,
            ).unsafe_ask()

            if launch_type == "Remote pipeline":
                try:
                    self.pipeline = nf_core.utils.prompt_remote_pipeline_name(self.wfs)
                except AssertionError as e:
                    log.error(e.args[0])
                    return False
            else:
                self.pipeline = questionary.path(
                    "Path to workflow:", style=nf_core.utils.nfcore_question_style
                ).unsafe_ask()

        # Get the schema
        self.schema_obj = PipelineSchema()
        if self.schema_obj is None:
            return False
        self.schema_obj.get_schema_path(self.pipeline, local_only=False, revision=self.pipeline_revision)
        self.schema_obj.get_wf_params()
        return True

    def format_group(self, definition, show_hidden=False) -> str:
        """Format a group of parameters of the schema as commented YAML.

        Args:
            definition (dict): Definition of the group from the schema
            show_hidden (bool): Whether to include hidden parameters

        Returns:
            str: Formatted output for a group
        """
        out = ""
        title = definition.get("title", definition)
        description = definition.get("description", "")
        properties = definition.get("properties")

        hidden_props = set()
        for param_key, param_props in properties.items():
            if param_props.get("hidden", False):
                hidden_props.add(param_key)

        out += _print_wrapped(title, "=", mode="both")
        if description:
            out += _print_wrapped(description, mode="none")

        out += "\n"
        if len(hidden_props) > 0:
            out += _print_wrapped(f"({len(hidden_props)} hidden parameters are not shown)", mode="none")
            out += "\n\n"

        required_props = definition.get("required", [])

        for prop_key, props in properties.items():
            param_out = self.format_param(prop_key, props, required_props, show_hidden=show_hidden)
            if param_out is not None:
                out += param_out
                out += "\n"

        return out

    def format_param(
        self, name: str, properties: Dict, required_properties: List[str] = [], show_hidden: bool = False
    ) -> Optional[str]:
        """
        Format a single parameter of the schema as commented YAML

        Args:
            name (str): Name of the parameter
            properties (dict): Properties of the parameter
            required_properties (list): List of required properties
            show_hidden (bool): Whether to include hidden parameters

        Returns:
            str: Section of a params-file.yml for given parameter
            None: If the parameter is skipped because it is hidden and show_hidden is not set
        """
        out = ""
        hidden = properties.get("hidden", False)

        if not show_hidden and hidden:
            return None

        description = properties.get("description", "")
        if self.schema_obj is None:
            log.error("No schema object found")
            return ""
        self.schema_obj.get_schema_defaults()
        default = properties.get("default")
        type = properties.get("type")
        required = name in required_properties

        out += _print_wrapped(name, "-", mode="both")

        if description:
            out += _print_wrapped(description + "\n", mode="none", indent=4)

        if type:
            out += _print_wrapped(f"Type: {type}", mode="none", indent=4)

        if required:
            out += _print_wrapped("Required", mode="none", indent=4)

        out += _print_wrapped("\n", mode="end")
        out += f"# {name} = {json.dumps(default)}\n"

        return out

    def generate_params_file(self, show_hidden: bool = False) -> str:
        """Generate the contents of a parameter template file.

        Assumes the pipeline has been fetched (if remote) and the schema loaded.

        Args:
            show_hidden (bool): Whether to include hidden parameters

        Returns:
            str: Formatted output for the pipeline schema
        """
        if self.schema_obj is None:
            log.error("No schema object found")
            return ""

        schema = self.schema_obj.schema
        pipeline_name = self.schema_obj.pipeline_manifest.get("name", self.pipeline)
        pipeline_version = self.schema_obj.pipeline_manifest.get("version", "0.0.0")

        # Build the header section
        out = ""

        out += _print_wrapped(f"{pipeline_name} {pipeline_version}", "~", mode="both", indent=4)
        out += _print_wrapped(INTRO.format(pipeline_name=pipeline_name), " ", mode="none", indent=4)
        out += _print_wrapped("\n", " ", mode="none", indent=4, drop_whitespace=False)
        out += _print_wrapped(USAGE, "-", mode="end", indent=4)
        out += "\n"

        # Add all parameter groups
        for definition in schema.get("definitions", schema.get("$defs", {})).values():
            out += self.format_group(definition, show_hidden=show_hidden)
            out += "\n"

        return out

    def write_params_file(self, output_fn: Path = Path("nf-params.yaml"), show_hidden=False, force=False) -> bool:
        """Build a template file for the pipeline schema.

        Args:
            output_fn (str, optional): Filename to write the template to.
            show_hidden (bool, optional):
                Include parameters marked as hidden in the output
            force (bool, optional): Whether to overwrite existing output file.

        Returns:
            bool: True if the template was written successfully, False otherwise
        """

        self.get_pipeline()
        if self.schema_obj is None:
            log.error("No schema object found")
            return False
        try:
            self.schema_obj.load_schema()
            self.schema_obj.validate_schema()
        except AssertionError as e:
            log.error(f'Pipeline schema file is invalid ("{self.schema_obj.schema_filename}"): {e}')
            log.info("Please fix this file, then try again.")
            return False

        schema_out = self.generate_params_file(show_hidden=show_hidden)

        if output_fn.exists() and not force:
            log.error(f"File '{output_fn}' exists! Please delete first, or use '--force'")
            return False
        output_fn.write_text(schema_out)
        log.info(f"Parameter file written to '{output_fn}'")

        return True



import logging
from pathlib import Path

log = logging.getLogger(__name__)


def system_exit(self):
    """Check for System.exit calls in groovy/nextflow code

    Calls to System.exit(1) should be replaced by throwing errors

    This lint test looks for all calls to `System.exit`
    in any file with the `.nf` or `.groovy` extension
    """
    passed = []
    warned = []

    root_dir = Path(self.wf_path)

    # Get all groovy and nf files
    groovy_files = [f for f in root_dir.rglob("*.groovy")]
    nf_files = [f for f in root_dir.rglob("*.nf")]
    to_check = nf_files + groovy_files

    for file in to_check:
        try:
            with file.open() as fh:
                for i, line in enumerate(fh.readlines(), start=1):
                    if "System.exit" in line and "System.exit(0)" not in line:
                        warned.append(f"`System.exit` in {file.name}: _{line.strip()}_  [line {i}]")
        except FileNotFoundError:
            log.debug(f"Could not open file {file.name} in system_exit lint test")

    if len(warned) == 0:
        passed.append("No `System.exit` calls found")

    return {"passed": passed, "warned": warned}



"""
Code for linting modules in the nf-core/modules repository and
in nf-core pipelines

Command:
nf-core modules lint
"""

import logging
import os
from pathlib import Path
from typing import List, Optional, Union

import questionary
import rich
import rich.progress
import ruamel.yaml

import nf_core.components
import nf_core.components.nfcore_component
import nf_core.modules.modules_utils
import nf_core.utils
from nf_core.components.components_utils import get_biotools_id
from nf_core.components.lint import ComponentLint, LintExceptionError, LintResult
from nf_core.components.nfcore_component import NFCoreComponent
from nf_core.pipelines.lint_utils import console, run_prettier_on_file

log = logging.getLogger(__name__)

from .environment_yml import environment_yml
from .main_nf import main_nf
from .meta_yml import meta_yml, obtain_correct_and_specified_inputs, obtain_correct_and_specified_outputs, read_meta_yml
from .module_changes import module_changes
from .module_deprecations import module_deprecations
from .module_patch import module_patch
from .module_tests import module_tests
from .module_todos import module_todos
from .module_version import module_version


class ModuleLint(ComponentLint):
    """
    An object for linting modules either in a clone of the 'nf-core/modules'
    repository or in any nf-core pipeline directory
    """

    # Import lint functions
    environment_yml = environment_yml
    main_nf = main_nf
    meta_yml = meta_yml
    obtain_correct_and_specified_inputs = obtain_correct_and_specified_inputs
    obtain_correct_and_specified_outputs = obtain_correct_and_specified_outputs
    read_meta_yml = read_meta_yml
    module_changes = module_changes
    module_deprecations = module_deprecations
    module_patch = module_patch
    module_tests = module_tests
    module_todos = module_todos
    module_version = module_version

    def __init__(
        self,
        directory: Union[str, Path],
        fail_warned: bool = False,
        fix: bool = False,
        remote_url: Optional[str] = None,
        branch: Optional[str] = None,
        no_pull: bool = False,
        registry: Optional[str] = None,
        hide_progress: bool = False,
    ):
        super().__init__(
            component_type="modules",
            directory=directory,
            fail_warned=fail_warned,
            fix=fix,
            remote_url=remote_url,
            branch=branch,
            no_pull=no_pull,
            registry=registry,
            hide_progress=hide_progress,
        )

    def lint(
        self,
        module=None,
        registry="quay.io",
        key=(),
        all_modules=False,
        print_results=True,
        show_passed=False,
        sort_by="test",
        local=False,
        fix_version=False,
    ):
        """
        Lint all or one specific module

        First gets a list of all local modules (in modules/local/process) and all modules
        installed from nf-core (in modules/nf-core)

        For all nf-core modules, the correct file structure is assured and important
        file content is verified. If directory subject to linting is a clone of 'nf-core/modules',
        the files necessary for testing the modules are also inspected.

        For all local modules, the '.nf' file is checked for some important flags, and warnings
        are issued if untypical content is found.

        :param module:          A specific module to lint
        :param print_results:   Whether to print the linting results
        :param show_passed:     Whether passed tests should be shown as well
        :param fix_version:     Update the module version if a newer version is available
        :param hide_progress:   Don't show progress bars

        :returns:               A ModuleLint object containing information of
                                the passed, warned and failed tests
        """
        # TODO: consider unifying modules and subworkflows lint() function and add it to the ComponentLint class
        # Prompt for module or all
        if module is None and not all_modules and len(self.all_remote_components) > 0:
            questions = [
                {
                    "type": "list",
                    "name": "all_modules",
                    "message": "Lint all modules or a single named module?",
                    "choices": ["All modules", "Named module"],
                },
                {
                    "type": "autocomplete",
                    "name": "tool_name",
                    "message": "Tool name:",
                    "when": lambda x: x["all_modules"] == "Named module",
                    "choices": [m.component_name for m in self.all_remote_components],
                },
            ]
            answers = questionary.unsafe_prompt(questions, style=nf_core.utils.nfcore_question_style)
            all_modules = answers["all_modules"] == "All modules"
            module = answers.get("tool_name")

        # Only lint the given module
        if module:
            if all_modules:
                raise LintExceptionError("You cannot specify a tool and request all tools to be linted.")
            local_modules = []
            remote_modules = [m for m in self.all_remote_components if m.component_name == module]
            if len(remote_modules) == 0:
                raise LintExceptionError(f"Could not find the specified module: '{module}'")
        else:
            local_modules = self.all_local_components
            remote_modules = self.all_remote_components

        if self.repo_type == "modules":
            log.info(f"Linting modules repo: [magenta]'{self.directory}'")
        else:
            log.info(f"Linting pipeline: [magenta]'{self.directory}'")
        if module:
            log.info(f"Linting module: [magenta]'{module}'")

        # Filter the tests by the key if one is supplied
        if key:
            self.filter_tests_by_key(key)
            log.info("Only running tests: '{}'".format("', '".join(key)))

        # If it is a pipeline, load the lint config file and the modules.json file
        if self.repo_type == "pipeline":
            self.set_up_pipeline_files()

        # Lint local modules
        if local and len(local_modules) > 0:
            self.lint_modules(local_modules, registry=registry, local=True, fix_version=fix_version)

        # Lint nf-core modules
        if len(remote_modules) > 0:
            self.lint_modules(remote_modules, registry=registry, local=False, fix_version=fix_version)

        if print_results:
            self._print_results(show_passed=show_passed, sort_by=sort_by)
            self.print_summary()

    def lint_modules(
        self, modules: List[NFCoreComponent], registry: str = "quay.io", local: bool = False, fix_version: bool = False
    ) -> None:
        """
        Lint a list of modules

        Args:
            modules ([NFCoreComponent]): A list of module objects
            registry (str): The container registry to use. Should be quay.io in most situations.
            local (boolean): Whether the list consist of local or nf-core modules
            fix_version (boolean): Fix the module version if a newer version is available
        """
        # TODO: consider unifying modules and subworkflows lint_modules() function and add it to the ComponentLint class
        progress_bar = rich.progress.Progress(
            "[bold blue]{task.description}",
            rich.progress.BarColumn(bar_width=None),
            "[magenta]{task.completed} of {task.total}[reset] » [bold yellow]{task.fields[test_name]}",
            transient=True,
            console=console,
            disable=self.hide_progress or os.environ.get("HIDE_PROGRESS", None) is not None,
        )
        with progress_bar:
            lint_progress = progress_bar.add_task(
                f"Linting {'local' if local else 'nf-core'} modules",
                total=len(modules),
                test_name=modules[0].component_name,
            )

            for mod in modules:
                progress_bar.update(lint_progress, advance=1, test_name=mod.component_name)
                self.lint_module(mod, progress_bar, local=local, fix_version=fix_version)

    def lint_module(
        self,
        mod: NFCoreComponent,
        progress_bar: rich.progress.Progress,
        local: bool = False,
        fix_version: bool = False,
    ):
        """
        Perform linting on one module

        If the module is a local module we only check the `main.nf` file,
        and issue warnings instead of failures.

        If the module is a nf-core module we check for existence of the files
        - main.nf
        - meta.yml
        And verify that their content conform to the nf-core standards.

        If the linting is run for modules in the central nf-core/modules repo
        (repo_type==modules), files that are relevant for module testing are
        also examined
        """
        # TODO: consider unifying modules and subworkflows lint_module() function and add it to the ComponentLint class
        # Only check the main script in case of a local module
        if local:
            self.main_nf(mod, fix_version, self.registry, progress_bar)
            self.passed += [LintResult(mod, *m) for m in mod.passed]
            warned = [LintResult(mod, *m) for m in (mod.warned + mod.failed)]
            if not self.fail_warned:
                self.warned += warned
            else:
                self.failed += warned

        # Otherwise run all the lint tests
        else:
            mod.get_inputs_from_main_nf()
            mod.get_outputs_from_main_nf()
            # Update meta.yml file if requested
            if self.fix:
                self.update_meta_yml_file(mod)

            if self.repo_type == "pipeline" and self.modules_json and mod.repo_url:
                # Set correct sha
                version = self.modules_json.get_module_version(mod.component_name, mod.repo_url, mod.org)
                mod.git_sha = version

            for test_name in self.lint_tests:
                if test_name == "main_nf":
                    getattr(self, test_name)(mod, fix_version, self.registry, progress_bar)
                else:
                    getattr(self, test_name)(mod)

            self.passed += [LintResult(mod, *m) for m in mod.passed]
            warned = [LintResult(mod, *m) for m in mod.warned]
            if not self.fail_warned:
                self.warned += warned
            else:
                self.failed += warned

            self.failed += [LintResult(mod, *m) for m in mod.failed]

    def update_meta_yml_file(self, mod):
        """
        Update the meta.yml file with the correct inputs and outputs
        """
        meta_yml = self.read_meta_yml(mod)
        corrected_meta_yml = meta_yml.copy()
        yaml = ruamel.yaml.YAML()
        yaml.preserve_quotes = True
        yaml.indent(mapping=2, sequence=2, offset=0)

        # Obtain inputs and outputs from main.nf and meta.yml
        # Used to compare only the structure of channels and elements
        # Do not compare features to allow for custom features in meta.yml (i.e. pattern)
        if "input" in meta_yml:
            correct_inputs, meta_inputs = self.obtain_correct_and_specified_inputs(mod, meta_yml)
        if "output" in meta_yml:
            correct_outputs, meta_outputs = self.obtain_correct_and_specified_outputs(mod, meta_yml)

        if "input" in meta_yml and correct_inputs != meta_inputs:
            log.debug(
                f"Correct inputs: '{correct_inputs}' differ from current inputs: '{meta_inputs}' in '{mod.meta_yml}'"
            )
            corrected_meta_yml["input"] = mod.inputs.copy()  # list of lists (channels) of dicts (elements)
            for i, channel in enumerate(corrected_meta_yml["input"]):
                for j, element in enumerate(channel):
                    element_name = list(element.keys())[0]
                    for k, meta_element in enumerate(meta_yml["input"]):
                        try:
                            # Handle old format of meta.yml: list of dicts (channels)
                            if element_name in meta_element.keys():
                                # Copy current features of that input element form meta.yml
                                for feature in meta_element[element_name].keys():
                                    if feature not in element[element_name].keys():
                                        corrected_meta_yml["input"][i][j][element_name][feature] = meta_element[
                                            element_name
                                        ][feature]
                                break
                        except AttributeError:
                            # Handle new format of meta.yml: list of lists (channels) of elements (dicts)
                            for x, meta_ch_element in enumerate(meta_element):
                                if element_name in meta_ch_element.keys():
                                    # Copy current features of that input element form meta.yml
                                    for feature in meta_element[x][element_name].keys():
                                        if feature not in element[element_name].keys():
                                            corrected_meta_yml["input"][i][j][element_name][feature] = meta_element[x][
                                                element_name
                                            ][feature]
                                    break

        if "output" in meta_yml and correct_outputs != meta_outputs:
            log.debug(
                f"Correct outputs: '{correct_outputs}' differ from current outputs: '{meta_outputs}' in '{mod.meta_yml}'"
            )
            corrected_meta_yml["output"] = mod.outputs.copy()  # list of dicts (channels) with list of dicts (elements)
            for i, channel in enumerate(corrected_meta_yml["output"]):
                ch_name = list(channel.keys())[0]
                for j, element in enumerate(channel[ch_name]):
                    element_name = list(element.keys())[0]
                    for k, meta_element in enumerate(meta_yml["output"]):
                        if element_name in meta_element.keys():
                            # Copy current features of that output element form meta.yml
                            for feature in meta_element[element_name].keys():
                                if feature not in element[element_name].keys():
                                    corrected_meta_yml["output"][i][ch_name][j][element_name][feature] = meta_element[
                                        element_name
                                    ][feature]
                            break
                        elif ch_name in meta_element.keys():
                            # When the previous output element was using the name of the channel
                            # Copy current features of that output element form meta.yml
                            try:
                                # Handle old format of meta.yml
                                for feature in meta_element[ch_name].keys():
                                    if feature not in element[element_name].keys():
                                        corrected_meta_yml["output"][i][ch_name][j][element_name][feature] = (
                                            meta_element[ch_name][feature]
                                        )
                            except AttributeError:
                                # Handle new format of meta.yml
                                for x, meta_ch_element in enumerate(meta_element[ch_name]):
                                    for meta_ch_element_name in meta_ch_element.keys():
                                        for feature in meta_ch_element[meta_ch_element_name].keys():
                                            if feature not in element[element_name].keys():
                                                corrected_meta_yml["output"][i][ch_name][j][element_name][feature] = (
                                                    meta_ch_element[meta_ch_element_name][feature]
                                                )
                            break

        # Add bio.tools identifier
        for i, tool in enumerate(corrected_meta_yml["tools"]):
            tool_name = list(tool.keys())[0]
            if "identifier" not in tool[tool_name]:
                corrected_meta_yml["tools"][i][tool_name]["identifier"] = get_biotools_id(
                    mod.component_name if "/" not in mod.component_name else mod.component_name.split("/")[0]
                )

        with open(mod.meta_yml, "w") as fh:
            log.info(f"Updating {mod.meta_yml}")
            yaml.dump(corrected_meta_yml, fh)
            run_prettier_on_file(fh.name)



"""Test Pipeline Create App"""

from unittest import mock

from nf_core.pipelines.create import PipelineCreateApp

INIT_FILE = "../../nf_core/pipelines/create/__init__.py"


async def test_app_bindings():
    """Test that the app bindings work."""
    app = PipelineCreateApp()
    async with app.run_test() as pilot:
        # Test pressing the D key
        assert app.dark
        await pilot.press("d")
        assert not app.dark
        await pilot.press("d")
        assert app.dark

        # Test pressing the Q key
        await pilot.press("q")
        assert app.return_code == 0


def test_welcome(snap_compare):
    """Test snapshot for the first screen in the app. The welcome screen."""
    assert snap_compare(INIT_FILE, terminal_size=(100, 50))


def test_choose_type(snap_compare):
    """Test snapshot for the choose_type screen.
    Steps to get to this screen:
        screen welcome > press start >
        screen choose_type
    """

    async def run_before(pilot) -> None:
        await pilot.click("#start")

    assert snap_compare(INIT_FILE, terminal_size=(100, 50), run_before=run_before)


def test_basic_details_nfcore(snap_compare):
    """Test snapshot for the basic_details screen of an nf-core pipeline.
    Steps to get to this screen:
        screen welcome > press start >
        screen choose_type > press nf-core >
        screen basic_details
    """

    async def run_before(pilot) -> None:
        await pilot.click("#start")
        await pilot.click("#type_nfcore")

    assert snap_compare(INIT_FILE, terminal_size=(100, 50), run_before=run_before)


def test_basic_details_custom(snap_compare):
    """Test snapshot for the basic_details screen of a custom pipeline.
    Steps to get to this screen:
        screen welcome > press start >
        screen choose_type > press custom >
        screen basic_details
    """

    async def run_before(pilot) -> None:
        await pilot.click("#start")
        await pilot.click("#type_custom")

    assert snap_compare(INIT_FILE, terminal_size=(100, 50), run_before=run_before)


def test_type_nfcore(snap_compare):
    """Test snapshot for the type_nfcore screen.
    Steps to get to this screen:
        screen welcome > press start >
        screen choose_type > press nf-core >
        screen basic_details > enter pipeline details > press next >
        screen type_nfcore
    """

    async def run_before(pilot) -> None:
        await pilot.click("#start")
        await pilot.click("#type_nfcore")
        await pilot.click("#name")
        await pilot.press("m", "y", "p", "i", "p", "e", "l", "i", "n", "e")
        await pilot.press("tab")
        await pilot.press("A", " ", "c", "o", "o", "l", " ", "d", "e", "s", "c", "r", "i", "p", "t", "i", "o", "n")
        await pilot.press("tab")
        await pilot.press("M", "e")
        await pilot.click("#next")

    assert snap_compare(INIT_FILE, terminal_size=(100, 50), run_before=run_before)


def test_type_nfcore_validation(snap_compare):
    """Test snapshot for the type_nfcore screen.
    Validation errors should appear when input fields are empty.
    Steps to get to this screen:
        screen welcome > press start >
        screen choose_type > press nf-core >
        screen basic_details > press next >
        ERRORS
    """

    async def run_before(pilot) -> None:
        await pilot.click("#start")
        await pilot.click("#type_nfcore")
        await pilot.click("#next")
        await pilot.pause()

    assert snap_compare(INIT_FILE, terminal_size=(100, 50), run_before=run_before)


def test_type_custom(snap_compare):
    """Test snapshot for the type_custom screen.
    Steps to get to this screen:
        screen welcome > press start >
        screen choose_type > press custom >
        screen basic_details > enter pipeline details > press next >
        screen type_custom
    """

    async def run_before(pilot) -> None:
        await pilot.click("#start")
        await pilot.click("#type_custom")
        await pilot.click("#name")
        await pilot.press("tab")
        await pilot.press("m", "y", "p", "i", "p", "e", "l", "i", "n", "e")
        await pilot.press("tab")
        await pilot.press("A", " ", "c", "o", "o", "l", " ", "d", "e", "s", "c", "r", "i", "p", "t", "i", "o", "n")
        await pilot.press("tab")
        await pilot.press("M", "e")
        await pilot.click("#next")

    assert snap_compare(INIT_FILE, terminal_size=(100, 50), run_before=run_before)


def test_final_details(snap_compare):
    """Test snapshot for the final_details screen.
    Steps to get to this screen:
        screen welcome > press start >
        screen choose_type > press nf-core >
        screen basic_details > enter pipeline details > press next >
        screen type_nfcore > press continue >
        screen final_details
    """

    async def run_before(pilot) -> None:
        await pilot.click("#start")
        await pilot.click("#type_nfcore")
        await pilot.click("#name")
        await pilot.press("m", "y", "p", "i", "p", "e", "l", "i", "n", "e")
        await pilot.press("tab")
        await pilot.press("A", " ", "c", "o", "o", "l", " ", "d", "e", "s", "c", "r", "i", "p", "t", "i", "o", "n")
        await pilot.press("tab")
        await pilot.press("M", "e")
        await pilot.click("#next")
        await pilot.click("#continue")

    assert snap_compare(INIT_FILE, terminal_size=(100, 50), run_before=run_before)


def test_customisation_help(snap_compare):
    """Test snapshot for the type_custom screen - showing help messages.
    Steps to get to this screen:
        screen welcome > press start >
        screen choose_type > press nf-core >
        screen basic_details > enter pipeline details > press next >
        screen type_custom > press Show more
    """

    async def run_before(pilot) -> None:
        await pilot.click("#start")
        await pilot.click("#type_custom")
        await pilot.click("#name")
        await pilot.press("tab")
        await pilot.press("m", "y", "p", "i", "p", "e", "l", "i", "n", "e")
        await pilot.press("tab")
        await pilot.press("A", " ", "c", "o", "o", "l", " ", "d", "e", "s", "c", "r", "i", "p", "t", "i", "o", "n")
        await pilot.press("tab")
        await pilot.press("M", "e")
        await pilot.click("#next")
        await pilot.click("#igenomes")
        await pilot.press("tab")
        await pilot.press("enter")

    assert snap_compare(INIT_FILE, terminal_size=(100, 50), run_before=run_before)


def test_github_question(tmpdir, snap_compare):
    """Test snapshot for the github_repo_question screen.
    Steps to get to this screen:
        screen welcome > press start >
        screen choose_type > press nf-core >
        screen basic_details > enter pipeline details > press next >
        screen type_nfcore > press continue >
        screen final_details > press finish > close logging screen >
        screen github_repo_question
    """

    async def run_before(pilot) -> None:
        await pilot.click("#start")
        await pilot.click("#type_nfcore")
        await pilot.click("#name")
        await pilot.press("m", "y", "p", "i", "p", "e", "l", "i", "n", "e")
        await pilot.press("tab")
        await pilot.press("A", " ", "c", "o", "o", "l", " ", "d", "e", "s", "c", "r", "i", "p", "t", "i", "o", "n")
        await pilot.press("tab")
        await pilot.press("M", "e")
        await pilot.click("#next")
        await pilot.click("#continue")
        await pilot.press("backspace")
        await pilot.press("tab")
        await pilot.press(*str(tmpdir))
        await pilot.click("#finish")
        await pilot.app.workers.wait_for_complete()
        await pilot.click("#close_screen")

    assert snap_compare(INIT_FILE, terminal_size=(100, 50), run_before=run_before)


@mock.patch("nf_core.pipelines.create.githubrepo.GithubRepo._get_github_credentials")
def test_github_details(mock_get_github_credentials, tmpdir, snap_compare):
    """Test snapshot for the github_repo screen.
    Steps to get to this screen:
        screen welcome > press start >
        screen choose_type > press nf-core >
        screen basic_details > enter pipeline details > press next >
        screen type_nfcore > press continue >
        screen final_details > press finish > close logging screen >
        screen github_repo_question > press create repo >
        screen github_repo
    """

    async def run_before(pilot) -> None:
        mock_get_github_credentials.return_value = (
            None,
            None,
        )  # mock the github credentials to have consistent snapshots
        await pilot.click("#start")
        await pilot.click("#type_nfcore")
        await pilot.click("#name")
        await pilot.press("m", "y", "p", "i", "p", "e", "l", "i", "n", "e")
        await pilot.press("tab")
        await pilot.press("A", " ", "c", "o", "o", "l", " ", "d", "e", "s", "c", "r", "i", "p", "t", "i", "o", "n")
        await pilot.press("tab")
        await pilot.press("M", "e")
        await pilot.click("#next")
        await pilot.click("#continue")
        await pilot.press("backspace")
        await pilot.press("tab")
        await pilot.press(*str(tmpdir))
        await pilot.click("#finish")
        await pilot.app.workers.wait_for_complete()
        await pilot.click("#close_screen")
        await pilot.click("#github_repo")

    assert snap_compare(INIT_FILE, terminal_size=(100, 50), run_before=run_before)


def test_github_exit_message(tmpdir, snap_compare):
    """Test snapshot for the github_exit screen.
    Steps to get to this screen:
        screen welcome > press start >
        screen choose_type > press nf-core >
        screen basic_details > enter pipeline details > press next >
        screen type_nfcore > press continue >
        screen final_details > press finish > close logging screen >
        screen github_repo_question > press create repo >
        screen github_repo > press exit (close without creating a repo) >
        screen github_exit
    """

    async def run_before(pilot) -> None:
        await pilot.click("#start")
        await pilot.click("#type_nfcore")
        await pilot.click("#name")
        await pilot.press("m", "y", "p", "i", "p", "e", "l", "i", "n", "e")
        await pilot.press("tab")
        await pilot.press("A", " ", "c", "o", "o", "l", " ", "d", "e", "s", "c", "r", "i", "p", "t", "i", "o", "n")
        await pilot.press("tab")
        await pilot.press("M", "e")
        await pilot.click("#next")
        await pilot.click("#continue")
        await pilot.press("backspace")
        await pilot.press("tab")
        await pilot.press(*str(tmpdir))
        await pilot.click("#finish")
        await pilot.app.workers.wait_for_complete()
        await pilot.click("#close_screen")
        await pilot.click("#github_repo")
        await pilot.click("#exit")

    assert snap_compare(INIT_FILE, terminal_size=(100, 50), run_before=run_before)



"""
The ComponentsTest class handles the generation and testing of nf-test snapshots.
"""

import logging
import os
import re
from pathlib import Path
from typing import List, Optional

import questionary
from rich import print
from rich.panel import Panel
from rich.prompt import Confirm
from rich.syntax import Syntax
from rich.text import Text

import nf_core.utils
from nf_core.components.components_command import ComponentCommand

log = logging.getLogger(__name__)


class ComponentsTest(ComponentCommand):  # type: ignore[misc]
    """
    Class to generate and test nf-test snapshots for modules.

    ...

    Attributes
    ----------
    component_type : str
        type of component to test (modules or subworkflows)
    component_name : str
        name of the tool to run tests for
    directory: str
        path to modules repository directory
    no_prompts : bool
        flat indicating if prompts are used
    remote_url : str
        URL of the remote repository
    branch : str
        branch of the remote repository
    verbose : bool
        flag indicating if verbose output should be used
    update : bool
        flag indicating if the existing snapshot should be updated
    once : bool
        flag indicating if the test should be run only once
    profile : str
        container software to use (docker, singularity or conda)

    Methods
    -------
    run():
        Run test steps
    check_inputs():
        Check inputs. Ask for component_name if not provided and check that the directory exists
    generate_snapshot():
        Generate the nf-test snapshot using `nf-test test` command
    check_snapshot_stability():
        Run the nf-test twice and check if the snapshot changes
    """

    def __init__(
        self,
        component_type: str,
        component_name: Optional[str] = None,
        directory: str = ".",
        no_prompts: bool = False,
        remote_url: Optional[str] = None,
        branch: Optional[str] = None,
        verbose: bool = False,
        update: bool = False,
        once: bool = False,
        profile: Optional[str] = None,
    ):
        super().__init__(component_type, directory, remote_url, branch, no_prompts=no_prompts)
        self.component_name = component_name
        self.remote_url = remote_url
        self.branch = branch
        self.errors: List[str] = []
        self.verbose = verbose
        self.obsolete_snapshots: bool = False
        self.update = update
        self.once = once
        self.profile = profile

    def run(self) -> None:
        """Run build steps"""
        self.check_inputs()
        os.environ["NFT_DIFF"] = "pdiff"  # set nf-test differ to pdiff to get a better diff output
        os.environ["NFT_DIFF_ARGS"] = (
            "--line-numbers --expand-tabs=2"  # taken from https://code.askimed.com/nf-test/docs/assertions/snapshots/#snapshot-differences
        )
        with nf_core.utils.set_wd(self.directory):
            self.check_snapshot_stability()
        if len(self.errors) > 0:
            errors = "\n - ".join(self.errors)
            raise UserWarning(f"Ran, but found errors:\n - {errors}")
        else:
            log.info("All tests passed!")

    def check_inputs(self) -> None:
        """Do more complex checks about supplied flags."""
        # Check modules directory structure
        if self.component_type == "modules":
            self.check_modules_structure()

        # Get the component name if not specified
        if self.component_name is None:
            if self.no_prompts:
                raise UserWarning(
                    f"{self.component_type[:-1].title()} name not provided and prompts deactivated. Please provide the {self.component_type[:-1]} name{' as TOOL/SUBTOOL or TOOL' if self.component_type == 'modules' else ''}."
                )
            else:
                try:
                    self.component_name = questionary.autocomplete(
                        "Tool name:" if self.component_type == "modules" else "Subworkflow name:",
                        choices=self.components_from_repo(self.org),
                        style=nf_core.utils.nfcore_question_style,
                    ).unsafe_ask()
                except LookupError:
                    raise

        self.component_dir = Path(self.component_type, self.modules_repo.repo_path, *self.component_name.split("/"))

        # First, sanity check that the module directory exists
        if not Path(self.directory, self.component_dir).is_dir():
            raise UserWarning(
                f"Cannot find directory '{self.component_dir}'.{' Should be TOOL/SUBTOOL or TOOL' if self.component_type == 'modules' else ''}"
            )

        # Check container software to use
        if os.environ.get("PROFILE") is None and self.profile is None:
            os.environ["PROFILE"] = ""
            if self.no_prompts:
                log.info(
                    "Setting environment variable '$PROFILE' to Docker as not set otherwise.\n"
                    "To use Singularity set 'export PROFILE=singularity' in your shell before running this command."
                )
                os.environ["PROFILE"] = "docker"
            else:
                question = {
                    "type": "list",
                    "name": "profile",
                    "message": "Choose container software to run the test with",
                    "choices": ["Docker", "Singularity", "Conda"],
                }
                answer = questionary.unsafe_prompt([question], style=nf_core.utils.nfcore_question_style)
                profile = answer["profile"].lower()
                os.environ["PROFILE"] = profile

    def display_nftest_output(self, nftest_out: bytes, nftest_err: bytes) -> None:
        nftest_output = Text.from_ansi(nftest_out.decode())
        print(Panel(nftest_output, title="nf-test output"))
        if nftest_err:
            syntax = Syntax(nftest_err.decode(), "diff", theme="ansi_dark")
            print(Panel(syntax, title="nf-test error"))
            if "Different Snapshot:" in nftest_err.decode():
                log.error("nf-test failed due to differences in the snapshots")
                # prompt to update snapshot
                if self.no_prompts:
                    log.info("Updating snapshot")
                    self.update = True
                elif self.update is None:
                    answer = Confirm.ask(
                        "[bold][blue]?[/] nf-test found differences in the snapshot. Do you want to update it?",
                        default=True,
                    )
                    if answer:
                        log.info("Updating snapshot")
                        self.update = True
                    else:
                        log.debug("Snapshot not updated")
                if self.update:
                    # update snapshot using nf-test --update-snapshot
                    self.generate_snapshot()

            else:
                self.errors.append("nf-test failed")

    def generate_snapshot(self) -> bool:
        """Generate the nf-test snapshot using `nf-test test` command

        returns True if the test was successful, False otherwise
        """

        log.debug("Running nf-test test")

        # set verbose flag if self.verbose is True
        verbose = "--verbose --debug" if self.verbose else ""
        update = "--update-snapshot" if self.update else ""
        self.update = False  # reset self.update to False to test if the new snapshot is stable
        tag = f"subworkflows/{self.component_name}" if self.component_type == "subworkflows" else self.component_name
        profile = self.profile if self.profile else os.environ["PROFILE"]

        result = nf_core.utils.run_cmd(
            "nf-test",
            f"test --tag {tag} --profile {profile} {verbose} {update}",
        )
        if result is not None:
            nftest_out, nftest_err = result
            self.display_nftest_output(nftest_out, nftest_err)
            # check if nftest_out contains obsolete snapshots
            pattern = r"Snapshot Summary:.*?(\d+)\s+obsolete"
            compiled_pattern = re.compile(pattern, re.DOTALL)  # re.DOTALL to allow . to match newlines
            obsolete_snapshots = compiled_pattern.search(nftest_out.decode())
            if obsolete_snapshots:
                self.obsolete_snapshots = True

            # check if nf-test was successful
            if "Assertion failed:" in nftest_out.decode():
                return False
            elif "no valid tests found." in nftest_out.decode():
                log.error("Test file 'main.nf.test' not found")
                self.errors.append("Test file 'main.nf.test' not found")
                return False
            else:
                log.debug("nf-test successful")
                return True
        else:
            log.error("nf-test failed")
            self.errors.append("nf-test failed")
            return False

    def check_snapshot_stability(self) -> bool:
        """Run the nf-test twice and check if the snapshot changes"""
        log.info("Generating nf-test snapshot")
        if not self.generate_snapshot():
            return False  # stop here if the first run failed
        elif self.once:
            return True  # stop here if the test should be run only once
        log.info("Generating nf-test snapshot again to check stability")
        if not self.generate_snapshot():
            log.error("nf-test snapshot is not stable")
            self.errors.append("nf-test snapshot is not stable")
            return False

        else:
            if self.obsolete_snapshots:
                # ask if the user wants to remove obsolete snapshots using nf-test --clean-snapshot
                if self.no_prompts or Confirm.ask(
                    "nf-test found obsolete snapshots. Do you want to remove them?", default=True
                ):
                    profile = self.profile if self.profile else os.environ["PROFILE"]
                    log.info("Removing obsolete snapshots")
                    nf_core.utils.run_cmd(
                        "nf-test",
                        f"test --tag {self.component_name} --profile {profile} --clean-snapshot",
                    )
                else:
                    log.debug("Obsolete snapshots not removed")
            return True



import shutil
from unittest import TestCase

import pytest

from nf_core.utils import Pipeline

from .utils import create_tmp_pipeline


class TestPipelines(TestCase):
    def setUp(self) -> None:
        """Create a new Pipeline for testing"""
        self.tmp_dir, self.template_dir, self.pipeline_name, self.pipeline_dir = create_tmp_pipeline()
        self.pipeline_obj = Pipeline(self.pipeline_dir)
        self.pipeline_obj._load()

    def tearDown(self) -> None:
        """Remove the test pipeline directory"""
        shutil.rmtree(self.tmp_dir)

    def _make_pipeline_copy(self):
        """Make a copy of the test pipeline that can be edited

        Returns: Path to new temp directory with pipeline"""
        new_pipeline = self.tmp_dir / "nf-core-testpipeline-copy"
        shutil.copytree(self.pipeline_dir, new_pipeline)
        return new_pipeline

    @pytest.fixture(autouse=True)
    def _use_caplog(self, caplog):
        self.caplog = caplog



import json
import logging
from pathlib import Path
from typing import Union

import ruamel.yaml
from jsonschema import exceptions, validators

from nf_core.components.components_differ import ComponentsDiffer
from nf_core.components.lint import ComponentLint, LintExceptionError
from nf_core.components.nfcore_component import NFCoreComponent

log = logging.getLogger(__name__)


def meta_yml(module_lint_object: ComponentLint, module: NFCoreComponent) -> None:
    """
    Lint a ``meta.yml`` file

    The lint test checks that the module has
    a ``meta.yml`` file and that it follows the
    JSON schema defined in the ``modules/meta-schema.json``
    file in the nf-core/modules repository.

    In addition it checks that the module name
    and module input is consistent between the
    ``meta.yml`` and the ``main.nf``.

    If the module has inputs or outputs, they are expected to be
    formatted as:

    .. code-block:: groovy

        tuple val(foo) path(bar)
        val foo
        path foo

    or permutations of the above.

    Args:
        module_lint_object (ComponentLint): The lint object for the module
        module (NFCoreComponent): The module to lint

    """

    # Check if we have a patch file, get original file in that case
    meta_yaml = read_meta_yml(module_lint_object, module)
    if module.is_patched and module_lint_object.modules_repo.repo_path is not None:
        lines = ComponentsDiffer.try_apply_patch(
            module.component_type,
            module.component_name,
            module_lint_object.modules_repo.repo_path,
            module.patch_path,
            Path(module.component_dir).relative_to(module.base_dir),
            reverse=True,
        ).get("meta.yml")
        if lines is not None:
            yaml = ruamel.yaml.YAML()
            meta_yaml = yaml.safe_load("".join(lines))
    if module.meta_yml is None:
        raise LintExceptionError("Module does not have a `meta.yml` file")
    if meta_yaml is None:
        module.failed.append(("meta_yml_exists", "Module `meta.yml` does not exist", module.meta_yml))
        return
    else:
        module.passed.append(("meta_yml_exists", "Module `meta.yml` exists", module.meta_yml))

    # Confirm that the meta.yml file is valid according to the JSON schema
    valid_meta_yml = False
    try:
        with open(Path(module_lint_object.modules_repo.local_repo_dir, "modules/meta-schema.json")) as fh:
            schema = json.load(fh)
        validators.validate(instance=meta_yaml, schema=schema)
        module.passed.append(("meta_yml_valid", "Module `meta.yml` is valid", module.meta_yml))
        valid_meta_yml = True
    except exceptions.ValidationError as e:
        hint = ""
        if len(e.path) > 0:
            hint = f"\nCheck the entry for `{e.path[0]}`."
        if e.message.startswith("None is not of type 'object'") and len(e.path) > 2:
            hint = f"\nCheck that the child entries of {str(e.path[0])+'.'+str(e.path[2])} are indented correctly."
        if e.schema and isinstance(e.schema, dict) and "message" in e.schema:
            e.message = e.schema["message"]
            incorrect_value = meta_yaml
            for key in e.path:
                incorrect_value = incorrect_value[key]

            hint = hint + f"\nThe current value is `{incorrect_value}`."
        module.failed.append(
            (
                "meta_yml_valid",
                f"The `meta.yml` of the module {module.component_name} is not valid: {e.message}.{hint}",
                module.meta_yml,
            )
        )

    # Confirm that all input and output channels are correctly specified
    if valid_meta_yml:
        # confirm that the name matches the process name in main.nf
        if meta_yaml["name"].upper() == module.process_name:
            module.passed.append(
                (
                    "meta_name",
                    "Correct name specified in `meta.yml`.",
                    module.meta_yml,
                )
            )
        else:
            module.failed.append(
                (
                    "meta_name",
                    f"Conflicting `process` name between meta.yml (`{meta_yaml['name']}`) and main.nf (`{module.process_name}`)",
                    module.meta_yml,
                )
            )
        # Check that inputs are specified in meta.yml
        if len(module.inputs) > 0 and "input" not in meta_yaml:
            module.failed.append(
                (
                    "meta_input",
                    "Inputs not specified in module `meta.yml`",
                    module.meta_yml,
                )
            )
        elif len(module.inputs) > 0:
            module.passed.append(
                (
                    "meta_input",
                    "Inputs specified in module `meta.yml`",
                    module.meta_yml,
                )
            )
        else:
            log.debug(f"No inputs specified in module `main.nf`: {module.component_name}")
        # Check that all inputs are correctly specified
        if "input" in meta_yaml:
            correct_inputs, meta_inputs = obtain_correct_and_specified_inputs(module_lint_object, module, meta_yaml)

            if correct_inputs == meta_inputs:
                module.passed.append(
                    (
                        "correct_meta_inputs",
                        "Correct inputs specified in module `meta.yml`",
                        module.meta_yml,
                    )
                )
            else:
                module.failed.append(
                    (
                        "correct_meta_inputs",
                        f"Module `meta.yml` does not match `main.nf`. Inputs should contain: {correct_inputs}\nRun `nf-core modules lint --fix` to update the `meta.yml` file.",
                        module.meta_yml,
                    )
                )

        # Check that outputs are specified in meta.yml
        if len(module.outputs) > 0 and "output" not in meta_yaml:
            module.failed.append(
                (
                    "meta_output",
                    "Outputs not specified in module `meta.yml`",
                    module.meta_yml,
                )
            )
        elif len(module.outputs) > 0:
            module.passed.append(
                (
                    "meta_output",
                    "Outputs specified in module `meta.yml`",
                    module.meta_yml,
                )
            )
        # Check that all outputs are correctly specified
        if "output" in meta_yaml:
            correct_outputs, meta_outputs = obtain_correct_and_specified_outputs(module_lint_object, module, meta_yaml)

            if correct_outputs == meta_outputs:
                module.passed.append(
                    (
                        "correct_meta_outputs",
                        "Correct outputs specified in module `meta.yml`",
                        module.meta_yml,
                    )
                )
            else:
                module.failed.append(
                    (
                        "correct_meta_outputs",
                        f"Module `meta.yml` does not match `main.nf`. Outputs should contain: {correct_outputs}\nRun `nf-core modules lint --fix` to update the `meta.yml` file.",
                        module.meta_yml,
                    )
                )


def read_meta_yml(module_lint_object: ComponentLint, module: NFCoreComponent) -> Union[dict, None]:
    """
    Read a `meta.yml` file and return it as a dictionary

    Args:
        module_lint_object (ComponentLint): The lint object for the module
        module (NFCoreComponent): The module to read

    Returns:
        dict: The `meta.yml` file as a dictionary
    """
    meta_yaml = None
    yaml = ruamel.yaml.YAML()
    yaml.preserve_quotes = True
    # Check if we have a patch file, get original file in that case
    if module.is_patched:
        lines = ComponentsDiffer.try_apply_patch(
            module.component_type,
            module.component_name,
            module_lint_object.modules_repo.repo_path,
            module.patch_path,
            Path(module.component_dir).relative_to(module.base_dir),
            reverse=True,
        ).get("meta.yml")
        if lines is not None:
            meta_yaml = yaml.load("".join(lines))
    if meta_yaml is None:
        if module.meta_yml is None:
            return None
        with open(module.meta_yml) as fh:
            meta_yaml = yaml.load(fh)
    return meta_yaml


def obtain_correct_and_specified_inputs(_, module, meta_yaml):
    """
    Obtain the list of correct inputs and the elements of each input channel.

    Args:
        module (object): The module object.
        meta_yaml (dict): The meta.yml dictionary.

    Returns:
        tuple: A tuple containing two lists. The first list contains the correct inputs,
               and the second list contains the inputs specified in meta.yml.
    """
    correct_inputs = []
    for input_channel in module.inputs:
        channel_elements = []
        for element in input_channel:
            channel_elements.append(list(element.keys())[0])
        correct_inputs.append(channel_elements)

    meta_inputs = []
    for input_channel in meta_yaml["input"]:
        if isinstance(input_channel, list):  # Correct format
            channel_elements = []
            for element in input_channel:
                channel_elements.append(list(element.keys())[0])
            meta_inputs.append(channel_elements)
        elif isinstance(input_channel, dict):  # Old format
            meta_inputs.append(list(input_channel.keys())[0])

    return correct_inputs, meta_inputs


def obtain_correct_and_specified_outputs(_, module, meta_yaml):
    """
    Obtain the dictionary of correct outputs and elements of each output channel.

    Args:
        module (object): The module object.
        meta_yaml (dict): The meta.yml dictionary.

    Returns:
        correct_outputs (dict): A dictionary containing the correct outputs and their elements.
        meta_outputs (dict): A dictionary containing the outputs specified in meta.yml.
    """
    correct_outputs = {}
    for output_channel in module.outputs:
        channel_name = list(output_channel.keys())[0]
        channel_elements = []
        for element in output_channel[channel_name]:
            channel_elements.append(list(element.keys())[0])
        correct_outputs[channel_name] = channel_elements

    meta_outputs = {}
    for output_channel in meta_yaml["output"]:
        channel_name = list(output_channel.keys())[0]
        if isinstance(output_channel[channel_name], list):  # Correct format
            channel_elements = []
            for element in output_channel[channel_name]:
                channel_elements.append(list(element.keys())[0])
            meta_outputs[channel_name] = channel_elements
        elif isinstance(output_channel[channel_name], dict):  # Old format
            meta_outputs[channel_name] = []

    return correct_outputs, meta_outputs



import ast
import glob
import logging
import re
from typing import Dict, List

log = logging.getLogger(__name__)


def plugin_includes(self) -> Dict[str, List[str]]:
    """Checks the include statements in the all *.nf files for plugin includes

    When nf-schema is used in an nf-core pipeline, the include statements of the plugin
    functions have to use nf-schema instead of nf-validation and vice versa
    """
    config_plugins = [plugin.split("@")[0] for plugin in ast.literal_eval(self.nf_config.get("plugins", "[]"))]
    validation_plugin = "nf-validation" if "nf-validation" in config_plugins else "nf-schema"

    passed: List[str] = []
    warned: List[str] = []
    failed: List[str] = []
    ignored: List[str] = []

    plugin_include_pattern = re.compile(r"^include\s*{[^}]+}\s*from\s*[\"']plugin/([^\"']+)[\"']\s*$", re.MULTILINE)
    workflow_files = [
        file for file in glob.glob(f"{self.wf_path}/**/*.nf", recursive=True) if not file.startswith("./modules/")
    ]
    test_passed = True
    for file in workflow_files:
        with open(file) as of:
            plugin_includes = re.findall(plugin_include_pattern, of.read())
            for include in plugin_includes:
                if include not in ["nf-validation", "nf-schema"]:
                    continue
                if include != validation_plugin:
                    test_passed = False
                    failed.append(
                        f"Found a `{include}` plugin import in `{file[2:]}`, but `{validation_plugin}` was used in `nextflow.config`"
                    )

    if test_passed:
        passed.append("No wrong validation plugin imports have been found")

    return {"passed": passed, "warned": warned, "failed": failed, "ignored": ignored}



import json
import logging
import subprocess
from pathlib import Path
from typing import List, Union

import rich
import yaml
from rich.console import Console
from rich.table import Table

import nf_core.utils
from nf_core.utils import plural_s as _s

log = logging.getLogger(__name__)

# Create a console used by all lint tests
console = Console(force_terminal=nf_core.utils.rich_force_colors())


def print_joint_summary(lint_obj, module_lint_obj, subworkflow_lint_obj):
    """Print a joint summary of the general pipe lint tests and the module and subworkflow lint tests"""
    swf_passed = 0
    swf_warned = 0
    swf_failed = 0
    module_passed = 0
    module_warned = 0
    module_failed = 0
    if subworkflow_lint_obj is not None:
        swf_passed = len(subworkflow_lint_obj.passed)
        swf_warned = len(subworkflow_lint_obj.warned)
        swf_failed = len(subworkflow_lint_obj.failed)
    if module_lint_obj is not None:
        module_passed = len(module_lint_obj.passed)
        module_warned = len(module_lint_obj.warned)
        module_failed = len(module_lint_obj.failed)
    nbr_passed = len(lint_obj.passed) + module_passed + swf_passed
    nbr_ignored = len(lint_obj.ignored)
    nbr_fixed = len(lint_obj.fixed)
    nbr_warned = len(lint_obj.warned) + module_warned + swf_warned
    nbr_failed = len(lint_obj.failed) + module_failed + swf_failed

    summary_colour = "red" if nbr_failed > 0 else "green"
    table = Table(box=rich.box.ROUNDED, style=summary_colour)
    table.add_column("LINT RESULTS SUMMARY", no_wrap=True)
    table.add_row(rf"[green][✔] {nbr_passed:>3} Test{_s(nbr_passed)} Passed")
    if nbr_fixed:
        table.add_row(rf"[bright blue][?] {nbr_fixed:>3} Test{_s(nbr_fixed)} Fixed")
    table.add_row(rf"[grey58][?] {nbr_ignored:>3} Test{_s(nbr_ignored)} Ignored")
    table.add_row(rf"[yellow][!] {nbr_warned:>3} Test Warning{_s(nbr_warned)}")
    table.add_row(rf"[red][✗] {nbr_failed:>3} Test{_s(nbr_failed)} Failed")
    console.print(table)


def print_fixes(lint_obj):
    """Prints available and applied fixes"""

    if lint_obj.could_fix:
        fix_flags = "".join([f" --fix {fix}" for fix in lint_obj.could_fix])
        wf_dir = "" if lint_obj.wf_path == "." else f"--dir {lint_obj.wf_path}"
        fix_cmd = f"nf-core pipelines lint {wf_dir} {fix_flags}"
        console.print(
            "\nTip: Some of these linting errors can automatically be resolved with the following command:\n\n"
            f"[blue]    {fix_cmd}\n"
        )
    if len(lint_obj.fix):
        console.print(
            "Automatic fixes applied. "
            "Please check with 'git diff' and revert any changes you do not want with 'git checkout <file>'."
        )


def run_prettier_on_file(file: Union[Path, str, List[str]]) -> None:
    """Run the pre-commit hook prettier on a file.

    Args:
        file (Path | str): A file identifier as a string or pathlib.Path.

    Warns:
        If Prettier is not installed, a warning is logged.
    """

    nf_core_pre_commit_config = Path(nf_core.__file__).parent / ".pre-commit-prettier-config.yaml"
    args = ["pre-commit", "run", "--config", str(nf_core_pre_commit_config), "prettier"]
    if isinstance(file, List):
        args.extend(["--files", *file])
    else:
        args.extend(["--files", str(file)])

    try:
        subprocess.run(args, capture_output=True, check=True)
        log.debug(f"${subprocess.STDOUT}")
    except subprocess.CalledProcessError as e:
        if ": SyntaxError: " in e.stdout.decode():
            log.critical(f"Can't format {file} because it has a syntax error.\n{e.stdout.decode()}")
        elif "files were modified by this hook" in e.stdout.decode():
            all_lines = [line for line in e.stdout.decode().split("\n")]
            files = "\n".join(all_lines[3:])
            log.debug(f"The following files were modified by prettier:\n {files}")
        else:
            log.warning(
                "There was an error running the prettier pre-commit hook.\n"
                f"STDOUT: {e.stdout.decode()}\nSTDERR: {e.stderr.decode()}"
            )


def dump_json_with_prettier(file_name, file_content):
    """Dump a JSON file and run prettier on it.
    Args:
        file_name (Path | str): A file identifier as a string or pathlib.Path.
        file_content (dict): Content to dump into the JSON file
    """
    with open(file_name, "w") as fh:
        json.dump(file_content, fh, indent=4)
    run_prettier_on_file(file_name)


def dump_yaml_with_prettier(file_name: Union[Path, str], file_content: dict) -> None:
    """Dump a YAML file and run prettier on it.

    Args:
        file_name (Path | str): A file identifier as a string or pathlib.Path.
        file_content (dict): Content to dump into the YAML file
    """
    with open(file_name, "w") as fh:
        yaml.safe_dump(file_content, fh)
    run_prettier_on_file(file_name)


def ignore_file(lint_name: str, file_path: Path, dir_path: Path) -> List[List[str]]:
    """Ignore a file and add the result to the ignored list. Return the passed, failed, ignored and ignore_configs lists."""

    passed: List[str] = []
    failed: List[str] = []
    ignored: List[str] = []
    _, pipeline_conf = nf_core.utils.load_tools_config(dir_path)
    lint_conf = getattr(pipeline_conf, "lint", None) or None

    if lint_conf is None:
        ignore_entry: List[str] = []
    else:
        ignore_entry = lint_conf.get(lint_name, [])
    full_path = dir_path / file_path
    # Return a failed status if we can't find the file
    if not full_path.is_file():
        if isinstance(ignore_entry, bool) and not ignore_entry:
            ignored.append(f"`{file_path}` not found, but it is ignored.")
            ignore_entry = []
        else:
            failed.append(f"`{file_path}` not found.")
    else:
        passed.append(f"`{file_path}` found and not ignored.")

    # we handled the only case where ignore_entry should be a bool, convert it to a list, to make downstream code easier
    if isinstance(ignore_entry, bool):
        ignore_entry = []

    return [passed, failed, ignored, ignore_entry]
